# pytransport._General - general python scripts / tools
# Version 1.0
# W. Shields and J. Snuverink
# william.shields.2010@live.rhul.ac.uk

"""
General utilities for day to day housekeeping

Classes:
_Writer - a class used for writing any output during conversion.

"""
import numpy as _np
from scipy import constants as _con
import sys as _sys
import os as _os
import string as _string
import glob as _glob

import Reader as _Reader
from Data import _beamprops
from Data import ConversionData


class _Writer:
    """
    Class for writing terminal output, logfile output, and the converted machine. 
    Class is designed specifically for use with pytransport conversion. Class is 
    protected as it should be hidden.

    kwargs:
    debugOutput: bool, default = False. 
    If true, strings supplied to class functions for debug output will be written.
    writeToLog: bool, default = False. 
    If true, strings supplied to class functions will be written to a logfile.
    logfile: string, default = ''. Log file name.
    """
    def __init__(self, debugOutput=False, writeToLog=False, logfile=''):
        self.debug = debugOutput
        self.logfile = logfile
        self.outlog = writeToLog

    def Printout(self, line, outToTerminal=True):
        """
        Print line output string. Prints to output log if specified at class instantiation.
        Argument outToTerminal (bool, default = True) will print line to the terminal if True.
        """
        if outToTerminal:
            _sys.stdout.write(line+'\n')
        if self.outlog:
            if self.logfile == '':
                raise IOError("Invalid log file name: ''")
            logfile = open(self.logfile, 'a')
            logfile.write(line)
            logfile.write('\n')
            logfile.close()

    def DebugPrintout(self, line):
        """
        Print line debug output string to logfile only.
        """
        if self.debug:
            self.Printout(line, outToTerminal=False)

    def BeamDebugPrintout(self, beam, units):
        """
        Print the debug output strings for the beam definition.
        The beam must be a pytransport.Data._beamprops instance, and the units must be the
        pytransport.Data.ConversionData.units dict.
        """
        if not isinstance(beam, _beamprops):
            raise TypeError("Beam must be pytransport.Data._beamprops instance.")
        self.DebugPrintout('\tBeam definition :')
        self.DebugPrintout('\tdistrType = ' + beam.distrType)
        self.DebugPrintout('\tenergy = '  + _np.str(beam.tot_energy) + ' GeV')
        self.DebugPrintout('\tSigmaX = '  + _np.str(beam.SigmaX) + ' ' + units['x'])
        self.DebugPrintout('\tSigmaXP = ' + _np.str(beam.SigmaXP) + ' ' + units['xp'])
        self.DebugPrintout('\tSigmaY = '  + _np.str(beam.SigmaY) + ' ' + units['y'])
        self.DebugPrintout('\tSigmaYP = ' + _np.str(beam.SigmaYP) + ' ' + units['yp'])
        self.DebugPrintout('\tSigmaE = '  + _np.str(beam.SigmaE))
        self.DebugPrintout('\tSigmaT = '  + _np.str(beam.SigmaT))
        self.DebugPrintout('\t(Final brho = ' + _np.str(_np.round(beam.brho, 2)) + ' Tm)')
        self.DebugPrintout('\tTwiss Params:')
        self.DebugPrintout('\tBetaX = '  + _np.str(beam.betx) + ' ' + units['beta_func'])
        self.DebugPrintout('\tBetaY = '  + _np.str(beam.bety) + ' ' + units['beta_func'])
        self.DebugPrintout('\tAlphaX = ' + _np.str(beam.alfx))
        self.DebugPrintout('\tAlphaY = ' + _np.str(beam.alfy))
        self.DebugPrintout('\tEmittx = ' + _np.str(beam.emitx) + ' ' + units['emittance'])
        self.DebugPrintout('\tEmittY = ' + _np.str(beam.emity) + ' ' + units['emittance'])

    def ElementPrepDebugPrintout(self, elementType, numElements):
        """
        Print the debug output string as required in the element preparation stage.
        """
        debugString = "\tEntry is a " + elementType + ", adding to the element registry as element "
        debugString += numElements + "."
        self.DebugPrintout(debugString)

    def Write(self, convData, filename):
        """
        Write the converted TRANSPORT file to disk. A pytransport.Data.ConversionData 
        instance and filename (string)
        must be supplied.
        """
        if not isinstance(filename, _np.str):
            raise TypeError("Filename must be a string")
        if not isinstance(convData, ConversionData):
            raise TypeError("convData must be a pytransport.Data.ConversionData instance.")

        fname = ""
        directory = ""
        if convData.convprops.gmadoutput:
            fname = filename + '.gmad'
            directory = convData.convprops.gmadDir
        elif convData.convprops.madxoutput:
            fname = filename + '.madx'
            directory = convData.convprops.madxDir

        self.Printout('Writing to file: ' + directory + "/" + fname)
        if directory == "":
            convData.machine.Write(fname)
        else:
            if not CheckDirExists(directory):
                _os.mkdir(directory)
            _os.chdir(directory)
            convData.machine.Write(fname)
            _os.chdir('../')


def CheckDirExists(directory):
    dirs = _glob.glob('*/')
    if directory[-1] != '/':
        directory += '/'
    if directory in dirs:
        return True
    return False


def CheckIsAddition(line, filetype='input'):
    """
    Function to check if a BEAM line of TRANSPORT code is a beam definition or r.m.s addition.
    """
    # Output file is a standard format, any RMS addition line should always be 10 long.
    if filetype == 'output':
        if len(line) == 10:
            return True

    elif filetype == 'input':
        if len(line) > 8:
            if (line[8] == '0.') or (line[8] == '0'):
                return True
            else:
                return False
    else:
        raise ValueError("File type can only be input or output")


def CheckIsOutput(inputfile):
    """
    Function to check if a file is a standard TRANSPORT output file.
    Based upon existence of the lines::
        
         "0  XXX"
    
    being present, which represents the TRANSPORT indicator card line.
    X can be 0, 1, 2. Default is 0.
    """
    isOutput = False
    try:
        f = open(inputfile)
        for inputline in f:
            inputline = inputline.replace("\r", '')
            inputline = inputline.replace("\n", '')
            if inputline in _Reader._allowedIndicatorLines:
                isOutput = True
                break
        f.close()
    except IOError:
        raise IOError('Cannot open file.')
    return isOutput


def CheckIsSentinel(line):
    for element in line:
        if element[:8] == 'SENTINEL':
            return True
    return False


def CheckSingleLineOutputApplied(inputfile):
    """
    Function to check if the control element that print element output in
    a single line was successfully applied. Check needed as not all versions
    of TRANSPORT can run this type code.
    """
    _optics = _Reader._Optics()
    optics = _optics._getOptics(inputfile)

    for element in optics:
        if element == 'IO: UNDEFINED TYPE CODE 13. 19. ;':
            return True
    return False


def ConvertBunchLength(transport, bunch_length):
    """
    Function to convert bunch length unit in TRANSPORT into seconds.
    """
    scale = transport.scale[transport.units['bunch_length'][0]]
    blmeters = bunch_length * scale  # Bunch length scaled to metres
    blseconds = blmeters / (transport.beamprops.beta*_con.c)  # Length converted to seconds
    return blseconds


def FindEndOfLine(line):  # Find the end of the line of code
    endpos = -1
    breakloop = False
    if isinstance(line, _np.str):
        for charnum, char in enumerate(line):
            if char == ';':
                endpos = charnum
                break
    elif isinstance(line, _np.ndarray):
        for index, ele in enumerate(line):
            for char in ele:
                if char == ';':
                    endpos = index
                    breakloop = True
                    break
            if breakloop:
                break
    return endpos


def GetComment(line):
    """
    Function to extract a comment from a line.
    Will only find one comment per line.
    """
    concat = ''
    for ele in line:
        concat += ele
        concat += ' '
    commstart = _string.find(concat, '(')
    commend = _string.find(concat, ')')
    if commstart != -1 and commend != -1:
        comment = concat[commstart+1: commend]
        gmadcomment = '! '+comment
    else:
        gmadcomment = None
    return gmadcomment


def GetElementData(line):
    data = []
    for index, ele in enumerate(line[1:]):
        if ele != '':
            try:
                data.append(_np.float(ele))
            except ValueError:
                try:
                    data.append(_np.float(ele[0]))
                except ValueError:
                    pass
    return data


def GetFaceRotationAngles(data, linenum):

    def searchForAngle(linelist):
        angle = 0
        for line in linelist:
            try:
                elecode = _np.float(line[0])
                del elecode
            except ValueError:
                angle = 0
                break

            if _np.float(line[0]) == 4.0:
                break
            elif _np.float(line[0]) == 2.0:
                endof = FindEndOfLine(line[1])
                if endof != -1:
                    try:
                        angle = _np.round(_np.float(line[1][:endof]), 4)
                    except ValueError:
                        try:
                            angle = _np.round(_np.float(line[2][:endof]), 4)
                        except ValueError:
                            pass
                    else:
                        pass
                else:
                    try:
                        angle = _np.round(_np.float(line[1]), 4)
                    except ValueError:
                        try:
                            angle = _np.round(_np.float(line[2]), 4)
                        except ValueError:
                            pass
                    else:
                        pass
                break
            else:
                pass
        return angle

    lineList = [i for i in data[(linenum - 5):linenum]]
    lineList.reverse()  # Search for input poleface in reverse line order

    anglein = searchForAngle(lineList)
    angleout = searchForAngle(data[linenum + 1:(linenum + 6)])

    return anglein, angleout


def GetIndicator(data):
    """
    Function to read the indicator number. Must be 0, 1, or 2, where:
        0 is a new lattice
        1 is for fitting with the first lattice (from a 0 indicator file)
        2 if for a second fitting which suppresses the first fitting.
    """
    indc = 0
    linenum = 0
    for linenum, line in enumerate(data):
        if line[0] == '0':
            if line[1] == '\r' or '\n':
                indc = 0
                break
        if line[0] == '1':
            if line[1] == '\r' or '\n':
                indc = 1
                break
        if line[0] == '2':
            if line[1] == '\r' or '\n':
                indc = 2
                break
    return indc, linenum


def GetLabel(line):
    """
    Function to get element label from code line.
    """
    label = None
    for elenum, ele in enumerate(line):
        startslash = _string.find(ele, "/")
        startquote = _string.find(ele, "'")
        startequal = _string.find(ele, "=")
        startdbquote = _string.find(ele, '"')
        if startslash != -1:
            end = 1 + startslash + _string.find(ele[(startslash + 1):], "/")
            if end <= startslash:
                label = ele[startslash + 1:]
            else:
                label = ele[startslash + 1:end]
            break
        elif startquote != -1:
            end = 1 + startquote + _string.find(ele[(startslash + 1):], "'")
            if end <= startquote:
                label = ele[startquote + 1:]
            else:
                label = ele[startquote + 1:end]
            break
        elif startequal != -1:
            end = 1 + startequal + _string.find(ele[(startslash + 1):], "=")
            if end <= startequal:
                label = ele[startequal + 1:]
            else:
                label = ele[startequal + 1:end]
            break
        elif startdbquote != -1:
            end = 1 + startdbquote + _string.find(ele[(startdbquote + 1):], '"')
            if end <= startdbquote:
                label = ele[startdbquote + 1:]
            else:
                label = ele[startdbquote + 1:end]
            break
        else:
            label = None
    return label


def GetPreamble(data):  # Redundant until pybdsim can handle comments.
    """
    Function to read any preamble at the start of the TRANSPORT file.
    """
    # indc, linenum = GetIndicator(data)
    gmadpreamble = []
    # for line in self.Transport.data[:linenum-1]:
    for line in data:
        if line == '\r\n':
            pass
        else:
            gmadline = '!' + line
            gmadpreamble.append(gmadline)
    return gmadpreamble


def GetTypeNum(line):
    """
    Function to extract the element type number (type code).
    Written because element types can contain alphabetical
    characters when fits are used, e.g: 5.0A. Only the number
    is required, the use of fitting does not need to be known.
    """
    eleNum = line[0]
    characNum = 0
    if len(eleNum) > 2:
        for characNum in range(len(eleNum[2:])):
            try:
                converted = _np.float(eleNum[:characNum+2])
                del converted
            except ValueError:
                break
        typeNum = _np.float(eleNum[:characNum+2])
    else:
        typeNum = _np.float(eleNum)
    return typeNum


def JoinSplitLines(linenum, lattice):
    firstline = lattice[linenum].replace(';', '')
    latticeline = firstline  # Copy for later
    firstline = _np.array(firstline.split(' '), dtype=_np.str)
    firstline = RemoveIllegals(firstline)
    numericals = []
    nonnumericals = []
    # Keep entries that are strings of numbers
    for i in firstline:
        try:
            number = _np.float(i)
            numericals.append(_np.str(number))
        except ValueError:
            nonnumericals.append(i)

    # Number of numerical elements minus the first which should be the entry type number.
    # This is bascially a way of extracting any label or comments.
    numelements = len(numericals) - 1

    secline = lattice[linenum + 1].replace(';', '')
    secline = _np.array(secline.split(' '), dtype=_np.str)
    secline = RemoveIllegals(secline)
    secnumericals = []

    for i in secline:
        try:
            number = _np.float(i)
            secnumericals.append("%.4f" % number)
        except ValueError:
            pass

    # Second line should be 15 minus number of numerical elements from prev line.
    # This is done to skip erroneous numbers in the line such as '000' which have
    # appeared when lines have been split.
    secline = secnumericals[-15 + numelements:]
    numericals.extend(secline)

    # Add to latticeline so as to appear like one single line in the file
    seclinetxt = ""
    for i in secline:
        newline = "     " + i
        seclinetxt += newline
    latticeline += seclinetxt

    # Add name to output
    if len(nonnumericals) == 1:
        numericals.append(nonnumericals[0])
    line = _np.array(numericals)
    return latticeline, line


def OutputFitsToRegistry(transport, outputdata):
    isLegal = {'*DRIFT*': 3.0,
               '*QUAD*': 5.0,
               '*BEND*': 4.0}

    for line in outputdata:
        append = False
        linedict = {'elementnum': 0.0,
                    'name': '',
                    'length': 0.0}
        data = RemoveIllegals(line.split(' '))
        eledata = GetElementData(data)
        label = GetLabel(data)
        if data[0] in isLegal:
            linedict['elementnum'] = isLegal[data[0]]
            linedict['name'] = label
            linedict['data'] = eledata[1:]  # first value is elementnum.
            linedict['length'] = eledata[1]
            append = True

        # Only add an element with a name to the fitting registry.
        # (Element has to be named to be varied in the fitting routine).
        # Otherwise update the total length of the machine.
        if append and (label is not None) and (label != ''):
            transport.FitRegistry.AddToRegistry(linedict, line)
        else:
            transport.FitRegistry.UpdateLength(linedict)
    return transport


def ProcessFits(fits):  # redundant
    # First split the fitting output into its respective sections (input problem step).
    fitsections = []
    fitsstarts = []
    # Start line of each section
    for linenum, line in enumerate(fits):
        if line.startswith('1'):
            fitsstarts.append(linenum)

    for secnum in range(len(fitsstarts)):
        if secnum + 1 < len(fitsstarts):
            section = fits[fitsstarts[secnum]:fitsstarts[secnum + 1]]
        else:
            section = fits[fitsstarts[secnum]:]
        lines = []
        for line in section:
            lines.append(RemoveIllegals(line.split(' ')))
        fitsections.append(lines)

    magnetlines = []
    for section in fitsections:
        for line in section:
            if (len(line) > 0) and (line[0][0] == '*' and line[0][-1] == '*') and line[0] != '*FIT*':
                magnetlines.append(line)


def RemoveFileExt(inputfile):
    """
    Remove the file extension from the input file name. Only works on known extensions.
    """
    exts = [".txt", ".dat", ".DAT", ".TXT"]
    if inputfile[-4:] in exts:
        return inputfile[:-4]
    return inputfile


def RemoveIllegals(line):
    """
    Function to remove '' and stray characters from lines.
    """
    illegal = ['"', '', '(', ')']

    linelist = [element for element in line if element not in illegal]
    line = _np.array(linelist)
    return line


def RemoveLabel(line):
    """
    Function to remove the label from a line.
    """
    label, elenum = GetLabel(line)
    if label is not None:
        element = line[elenum]
        lablen = len(label)
        newval = element
        for index in range(len(element)):
            if element[index:index + lablen] == label:
                prelabel = element[:index - 1]
                postlabel = element[index + lablen + 1:]
                newval = prelabel + ' ' + postlabel
                break
        line[elenum] = newval
    return line


def RemoveSpaces(line):
    elementlist = []
    for i in line:
        if (i != '') and (i != ' '):
            elementlist.append(i)
    line = _np.array(elementlist)
    return line


def ScaleToMeters(transport, quantity):
    """
    Function to scale quantity (string) to meters, returns conversion factor.
    """
    if transport.units[quantity] != 'm':
        conversionFactor = transport.scale[transport.units[quantity][0]]
    else:
        conversionFactor = 1
    return conversionFactor


def UpdateEnergyFromMomentum(transport, momentum):
    """
    Function to calculate (from momentum):
        Total Energy
        Kinetic Energy
        Momentum
        Lorentz factor (gamma)
        Velocity (beta)
        Magnetic rigidity (brho)
    """
    momentum = _np.float(momentum)
    transport.beamprops.momentum = momentum
    p_mass = transport.beamprops.mass  # Particle rest mass (in GeV)
    scaling = 1
    mom_in_ev = momentum

    mom_unit = transport.units['p_egain']
    if mom_unit != 'eV':
        scaling = 1e9 / transport.scale[mom_unit[0]]     # Scaling relative to mom. unit
        mom_in_ev = momentum * transport.scale[mom_unit[0]]
    elif mom_unit == 'eV':
        scaling = 1e9                               # Scaling relative to 1 eV
        mom_in_ev = momentum
    p_mass *= scaling                               # Scale particle rest mass
    energy = _np.sqrt((p_mass**2) + (momentum**2))
    transport.beamprops.tot_energy = energy
    transport.beamprops.tot_energy_current = energy
    transport.beamprops.k_energy = energy - p_mass
    transport.beamprops.gamma = energy / p_mass
    transport.beamprops.beta = _np.sqrt((1.0 - (1.0 / transport.beamprops.gamma**2)))
    transport.beamprops.brho = mom_in_ev / _con.c
    return transport


def UpdateMomentumFromEnergy(transport, k_energy):
    """
    Function to calculate (from kinetic energy):
        Total Energy
        Kinetic Energy
        Momentum
        Lorentz factor (gamma)
        Velocity (beta)
        Magnetic rigidity (brho)
    """
    scaling = 1  # defaults
    mom_in_ev = 0
    k_energy = _np.float(k_energy)

    transport.beamprops.k_energy = k_energy
    p_mass = transport.beamprops.mass  # Particle rest mass (in GeV)

    e_unit = transport.units['p_egain']
    if e_unit != 'eV':
        scaling = 1e9 / transport.scale[e_unit[0]]  # Scaling relative to mom. unit
    elif e_unit == 'eV':
        scaling = 1e9  # Scaling relative to 1 eV
    p_mass *= scaling  # Scale particle rest mass

    # energy = _np.sqrt((p_mass**2 * _con.c**2) + (momentum**2 * _con.c**2)) / _con.c

    transport.beamprops.tot_energy_current = k_energy + p_mass
    transport.beamprops.momentum = _np.sqrt((transport.beamprops.tot_energy_current ** 2) - (p_mass ** 2))

    transport.beamprops.gamma = transport.beamprops.tot_energy_current / p_mass
    transport.beamprops.beta = _np.sqrt((1.0 - (1.0 / transport.beamprops.gamma ** 2)))

    if e_unit != 'eV':
        mom_in_ev = transport.beamprops.momentum * transport.scale[e_unit[0]]
    elif e_unit == 'eV':
        mom_in_ev = transport.beamprops.momentum

    transport.beamprops.brho = mom_in_ev / _con.c
    return transport



# pytransport.Data - Data loaders and data containers.
# Version 1.0
# W. Shields and J. Snuverink
# william.shields.2010@live.rhul.ac.uk

"""
Data

Data containers used in converting from Transport to gmad or madx.

Classes:
BDSData - a list of data read from Transport files.
ConversionData - a class for holding data during conversion.

"""

import numpy as _np
import os as _os
from scipy import constants as _con
import copy
from collections import defaultdict

_useRootNumpy = True

try:
    import root_numpy as _rnp
except (ImportError, AttributeError):
    _useRootNumpy = False
    pass


def _Load(filepath):
    extension = filepath.split('.')[-1]
    if not _os.path.isfile(filepath):
        raise IOError("File does not exist")
    elif extension == 'root':
        try:
            return _LoadRoot(filepath)
        except NameError:
            # raise error rather than return None, saves later scripting errors.
            raise IOError('Root loader not available.')
    else:
        raise IOError("Unknown file type - not BDSIM data")


def _LoadRoot(filepath):
    if not _useRootNumpy:
        raise IOError("root_numpy not available - can't load ROOT file")
    data = BDSData()
    trees = _rnp.list_trees(filepath)

    if 'optics' in trees:
        branches = _rnp.list_branches(filepath, 'optics')
        treedata = _rnp.root2array(filepath, 'optics')
    elif 'orbit' in trees:
        branches = _rnp.list_branches(filepath, 'orbit')
        treedata = _rnp.root2array(filepath, 'orbit')
    else:
        raise IOError("This file doesn't have the required tree 'optics'.")
    for element in range(len(treedata[branches[0]])):
        elementlist = []
        for branch in branches:
            if element == 0:
                data._AddProperty(branch)
            elementlist.append(treedata[branch][element])
        data.append(elementlist)
    return data


def _ParseHeaderLine(line):
    names = []
    units = []
    for word in line.split():
        if word.count('[') > 0:
            names.append(word.split('[')[0])
            units.append(word.split('[')[1].strip(']'))
        else:
            names.append(word)
            units.append('NA')
    return names, units


class BDSData(list):
    """
    General class representing simple 2 column data.

    Inherits python list.  It's a list of tuples with extra columns of 'name' and 'units'.
    """
    def __init__(self, *args, **kwargs):
        list.__init__(self, *args, **kwargs)
        self.units   = []
        self.names   = []
        self.columns = self.names

    def __getitem__(self, index):
        return dict(zip(self.names, list.__getitem__(self, index)))

    def GetItemTuple(self, index):
        """
        Get a specific entry in the data as a tuple of values rather than a dictionary.
        """
        return list.__getitem__(self, index)
        
    def _AddMethod(self, variablename):
        """
        This is used to dynamically add a getter function for a variable name.
        """
        def GetAttribute():
            if self.names.count(variablename) == 0:
                raise KeyError(variablename+" is not a variable in this data")
            ind = self.names.index(variablename)
            return _np.array([event[ind] for event in self])
        setattr(self, variablename, GetAttribute)

    def ConcatenateMachine(self, *args):
        """
        This is used to concatenate machines.
        """
        # Get final position of the machine (different param for survey)
        lastSpos = self.GetColumn('S')[-1]
        
        for machine in args:
            if isinstance(machine, _np.str):
                machine = _Load(machine)
        
            # check names sets are equal
            if len(set(self.names).difference(set(machine.names))) != 0:
                raise AttributeError("Cannot concatenate machine, variable names do not match")

            if self.names.count('S') != 0:
                sind = self.names.index('S')
            else:
                raise KeyError("S is not a variable in this data")
        
            # Have to convert each element to a list as tuples can't be modified
            for index in range(len(machine)):
                element = machine.GetItemTuple(index)
                elementlist = list(element)
                elementlist[sind] += lastSpos
                
                self.append(tuple(elementlist))

            lastSpos += machine.GetColumn('S')[-1]

    def _AddProperty(self, variablename, variableunit='NA'):
        """
        This is used to add a new variable and hence new getter function
        """
        self.names.append(variablename)
        self.units.append(variableunit)
        self._AddMethod(variablename)

    def _DuplicateNamesUnits(self, bdsdata2instance):
        d = bdsdata2instance
        for name, unit in zip(d.names, d.units):
            self._AddProperty(name, unit)

    def MatchValue(self, parametername, matchvalue, tolerance):
        """
        This is used to filter the instance of the class based on matching
        a parameter withing a certain tolerance.

        >>> a = pytransport.Data.Load("myfile.txt")
        >>> a.MatchValue("S",0.3,0.0004)
        
        this will match the "S" variable in instance "a" to the value of 0.3
        within +- 0.0004.

        You can therefore used to match any parameter.

        Return type is BDSAsciiData
        """
        if hasattr(self, parametername):
            a = BDSData()                 #build bdsdata2
            a._DuplicateNamesUnits(self)  #copy names and units
            pindex = a.names.index(parametername)
            filtereddata = [event for event in self if abs(event[pindex] - matchvalue) <= tolerance]
            a.extend(filtereddata)
            return a
        else:
            print "The parameter: ", parametername, " does not exist in this instance"

    def Filter(self, booleanarray):
        """
        Filter the data with a booleanarray.  Where true, will return
        that event in the data.

        Return type is BDSData
        """
        a = BDSData()
        a._DuplicateNamesUnits(self)
        a.extend([event for i, event in enumerate(self) if booleanarray[i]])
        return a

    def NameFromNearestS(self, S):
        i = self.IndexFromNearestS(S)
        if not hasattr(self, "Name"):
            raise ValueError("This file doesn't have the required column Name")
        return self.Name()[i]
    
    def IndexFromNearestS(self, S):
        """
        IndexFromNearestS(S) 

        return the index of the beamline element clostest to S 

        Only works if "SStart" column exists in data
        """
        # check this particular instance has the required columns for this function
        if not hasattr(self, "SStart"):
            raise ValueError("This file doesn't have the required column SStart")
        if not hasattr(self, "Arc_len"):
            raise ValueError("This file doesn't have the required column Arc_len")
        s = self.SStart()
        l = self.Arc_len()

        # iterate over beamline and record element if S is between the
        # sposition of that element and then next one
        # note madx S position is the end of the element by default
        ci = [i for i in range(len(self) - 1) if (S > s[i] and S < s[i]+l[i])]
        try:
            ci = ci[0] # return just the first match - should only be one
        except IndexError:
            # protect against S positions outside range of machine
            if S > s[-1]:
                ci -= 1
            else:
                ci = 0
        # check the absolute distance to each and return the closest one
        # make robust against s positions outside machine range
        return ci

    def GetColumn(self, columnstring):
        """
        Return a numpy array of the values in columnstring in order
        as they appear in the beamline
        """
        if columnstring not in self.columns:
            raise ValueError("Invalid column name")
        ind = self.names.index(columnstring)
        return _np.array([element[ind] for element in self])

    def __repr__(self):
        s = ''
        s += 'pytransport.Data.BDSData instance\n'
        s += str(len(self)) + ' entries'
        return s

    def MergeDuplicatesAtSameS(self):
        """
        Merge duplicate entries at the same s position. This is to prevent having multiple
        entries at the same s in situations such as poleface rotations. Will merge zero-length items
        into finite length items.
        """

        sPositions = defaultdict(list)
        for sIndex, sPos in enumerate(self.S()):
            sPositions[sPos].append(sIndex)
        duplicates = [(key, locs) for key, locs in sPositions.items() if len(locs) > 1]

        if len(duplicates):
            self.pop(duplicates[0][1][0])
            self.MergeDuplicatesAtSameS()


class ConversionData:
    """
    Class used as data container object in Transport2Gmad / Transport2Madx conversion.
    Required input:
    - inputfile: string, inputfile name
    - machine: either pybdsim.Builder.Machine or pymadx.Builder.Machine instance.

    Note: if used as a holder for conversion to gmad, options must be supplied a pybdsim.Options.Options instance.

    This class will hold ALL conversion related data, some stored in member variables which are separate containers
    for: conversion related properties (user input arguments), beam properties, and machine properties.
    """
    def __init__(self,
                 inputfile,
                 machine,
                 options       = None,  # None as madx has no options class. gmad needs pybdsim.Options passing in
                 particle      = 'proton',
                 debug         = False,
                 distrType     = 'gauss',
                 gmad          = True,
                 gmadDir       = 'gmad',
                 madx          = False,
                 madxDir       = 'madx',
                 auto          = True,
                 dontSplit     = False,
                 keepName      = False,
                 combineDrifts = False,
                 outlog        = True):

        if particle == 'proton':
            p_mass = _con.proton_mass * (_con.c ** 2 / _con.e) / 1e9  # Particle masses in same unit as TRANSPORT (GeV)
        elif particle == 'e-' or particle == 'e+':
            p_mass = _con.electron_mass * (_con.c ** 2 / _con.e) / 1e9
        else:
            p_mass = 1

        # pytransport data container classes. Split into different container classes so that this class is not
        # overloaded with member variables. Could be stored as a dictionary but given their widespread use in
        # conversion, access is just cleaner.
        self.convprops = _conversionProps(inputfile, particle, debug, gmad, gmadDir, madx, madxDir,
                                                   auto, dontSplit, keepName, combineDrifts, outlog)
        self.beamprops = _beamprops(p_mass)
        self.beamprops.distrType = distrType
        self.machineprops = _machineprops()

        # can't verify correct data type so at least check that it's not None if converting to gmad.
        if (options is None) and gmad:
            raise TypeError("options must be a pybdsim.Options.Options instance when converting to gmad")
        else:
            self.options = options

        # the gmad/madx machine and beam that will be written.
        self.machine = machine
        self.beam = self.machine.beam
        self.beam['offsetSampleMean'] = 0

        # make a copy of the empty machine. Copy needed in case machine is split and a new machine is needed.
        self._machineCopy = copy.deepcopy(self.machine)

        # initialise registries
        self.ElementRegistry = _Registry()
        self.FitRegistry = _Registry()

        self.units = {  # Default TRANSPORT units
            'x': 'cm',
            'xp': 'mrad',
            'y': 'cm',
            'yp': 'mrad',
            'bunch_length': 'cm',
            'momentum_spread': 'pc',
            'element_length': 'm',
            'magnetic_fields': 'kG',
            'p_egain': 'GeV',  # Momentum / energy gain during acceleration.
            'bend_vert_gap': 'cm',  # Vertical half-gap in dipoles
            'pipe_rad': 'cm',
            'beta_func': 'm',
            'emittance': 'mm mrad'
        }
        self.scale = {
            'p': 1e-12,
            'n': 1e-9,
            'u': 1e-6,
            'm': 1e-3,
            'c': 1e-2,
            'k': 1e+3,
            'K': 1e+3,  # Included both cases of k just in case.
            'M': 1e+6,
            'G': 1e+9,
            'T': 1e+12
        }

        self.accstart = []  # An index of the start of acceleration elements.
        self.data = []  # A list that will contain arrays of the element data
        self.filedata = []  # A list that will contain the raw strings from the input file

    def AddOptions(self):
        """
        Function to set the Options for a BDSIM machine.
        """
        self.options.SetPhysicsList(physicslist='em')
        self.options.SetBeamPipeRadius(beampiperadius=self.machineprops.beampiperadius,
                                       unitsstring=self.units['pipe_rad'])
        self.options.SetOuterDiameter(outerdiameter=0.5, unitsstring='m')
        self.options.SetTunnelRadius(tunnelradius=1, unitsstring='m')
        self.options.SetBeamPipeThickness(bpt=5, unitsstring='mm')
        self.options.SetSamplerDiameter(radius=1, unitsstring='m')
        self.options.SetStopSecondaries(stop=True)
        self.options.SetIncludeFringeFields(on=True)

        self.machine.AddOptions(self.options)

    def AddBeam(self):
        """
        Function to prepare the beam and add to the machine.
        """
        # convert energy to GeV (madx only handles GeV)
        energy_in_gev = self.beamprops.tot_energy * self.scale[self.units['p_egain'][0]] / 1e9
        self.beamprops.tot_energy = energy_in_gev

        self.beam.SetParticleType(self.convprops.particle)
        self.beam.SetEnergy(energy=self.beamprops.tot_energy, unitsstring='GeV')

        if self.convprops.gmadoutput:
            # set gmad parameters depending on distribution
            if self.beamprops.distrType == 'gausstwiss':
                self.beam.SetDistributionType(self.beamprops.distrType)
                self.beam.SetBetaX(self.beamprops.betx)
                self.beam.SetBetaY(self.beamprops.bety)
                self.beam.SetAlphaX(self.beamprops.alfx)
                self.beam.SetAlphaY(self.beamprops.alfy)
                self.beam.SetEmittanceX(self.beamprops.emitx, unitsstring='mm')
                self.beam.SetEmittanceY(self.beamprops.emity, unitsstring='mm')
                self.beam.SetSigmaE(self.beamprops.SigmaE)
                self.beam.SetSigmaT(self.beamprops.SigmaT)

            else:
                self.beam.SetDistributionType(self.beamprops.distrType)
                self.beam.SetSigmaX(self.beamprops.SigmaX, unitsstring=self.units['x'])
                self.beam.SetSigmaY(self.beamprops.SigmaY, unitsstring=self.units['y'])
                self.beam.SetSigmaXP(self.beamprops.SigmaXP, unitsstring=self.units['xp'])
                self.beam.SetSigmaYP(self.beamprops.SigmaYP, unitsstring=self.units['yp'])
                self.beam.SetSigmaE(self.beamprops.SigmaE)
                self.beam.SetSigmaT(self.beamprops.SigmaT)

            # set beam offsets in gmad if non zero
            if self.beamprops.X0 != 0:
                self.beam.SetX0(self.beamprops.X0, unitsstring=self.units['x'])
            if self.beamprops.Y0 != 0:
                self.beam.SetY0(self.beamprops.Y0, unitsstring=self.units['y'])
            if self.beamprops.Z0 != 0:
                self.beam.SetZ0(self.beamprops.Z0, unitsstring=self.units['z'])

        elif self.convprops.madxoutput:
            # calculate betas and emittances regardless for madx beam
            try:
                self.beamprops.betx = self.beamprops.SigmaX / self.beamprops.SigmaXP
            except ZeroDivisionError:
                self.beamprops.betx = 0
            try:
                self.beamprops.bety = self.beamprops.SigmaY / self.beamprops.SigmaYP
            except ZeroDivisionError:
                self.beamprops.bety = 0
                self.beamprops.emitx = self.beamprops.SigmaX * self.beamprops.SigmaXP / 1000.0
                self.beamprops.emity = self.beamprops.SigmaY * self.beamprops.SigmaYP / 1000.0

            # set madx beam
            self.beam.SetDistributionType('madx')
            self.beam.SetBetaX(self.beamprops.betx)
            self.beam.SetBetaY(self.beamprops.bety)
            self.beam.SetAlphaX(self.beamprops.alfx)
            self.beam.SetAlphaY(self.beamprops.alfy)
            self.beam.SetEmittanceX(self.beamprops.emitx / 1000)
            self.beam.SetEmittanceY(self.beamprops.emity / 1000)
            self.beam.SetSigmaE(self.beamprops.SigmaE)
            self.beam.SetSigmaT(self.beamprops.SigmaT)

        self.machine.AddBeam(self.beam)

    def ResetMachine(self):
        """
        Delete the machine and set to be the empty machine copied at class instantiation.
        """
        del self.machine
        self.machine = self._machineCopy


class _beamprops:
    """
    A class containing the properties of the beam distribution.
    """
    def __init__(self, p_mass=938.272):
        # beam properties that are updated along the lattice
        self.momentum = 0
        self.k_energy = 0
        self.tot_energy_current = p_mass
        self.gamma = 1
        self.beta = 0
        self.brho = 0
        # beam properties that are from the initial beam and fixed
        self.mass = p_mass
        self.tot_energy = p_mass  # initial energy
        self.SigmaX = 0
        self.SigmaY = 0
        self.SigmaXP = 0
        self.SigmaYP = 0
        self.SigmaE = 0
        self.SigmaT = 0
        self.X0 = 0
        self.Y0 = 0
        self.Z0 = 0
        self.T0 = 0
        self.Xp0 = 0
        self.Yp0 = 0
        self.betx = 0
        self.alfx = 0
        self.bety = 0
        self.alfy = 0
        self.dx = 0
        self.dy = 0
        self.emitx = 0
        self.emity = 0
        self.distrType = 'gauss'


class _machineprops:
    """
    A class containing the number of elements and angular properties (i.e bending direction)
    """
    def __init__(self):
        self.benddef          = True  # True = dipole defined by 4. L B n. False = dipole defined by 4. L angle n.
        self.bending          = 1     # +VE = bends to the right for positive particles
        self.angle            = 0     # dipole rotation angle
        self.drifts           = 0     # nr of drifts
        self.dipoles          = 0
        self.rf               = 0
        self.quads            = 0
        self.sextus           = 0
        self.transforms       = 0
        self.solenoids        = 0
        self.collimators      = 0
        self.beampiperadius   = 20
        self.fringeIntegral   = 0  # global value for all subsequent fringe fields until set otherwise
        self.secondfringeInt  = 0  # second fringe field integral
        self.bendInCurvature  = 0  # global value for all subsequent dipole entrance poleface curvatures
        self.bendOutCurvature = 0  # global value for all subsequent dipole exit poleface curvatures
        self.dipoleVertAper   = 0
        self.apertureType     = 'circular'
        self._totalAccVoltage = 0
        self._e_gain_prev     = 0


class _conversionProps:
    """
    A class containing the settings for the conversion.
    """
    def __init__(self, inputfile,
                 particle      = 'proton',
                 debug         = False,
                 gmad          = True,
                 gmadDir       = 'gmad',
                 madx          = False,
                 madxDir       = 'madx',
                 auto          = True,
                 dontSplit     = False,
                 keepName      = False,
                 combineDrifts = False,
                 outlog        = True):

        self.debug = debug
        self.outlog = outlog
        self.typeCode6IsTransUpdate = True  # Definition of type code 6, true is transform update, false is collimator
        self.isAccSequence = False  # Definition of type code 11 is not an accelerator sequence

        # Automatic writing and machine splitting
        self.auto = auto
        self.dontSplit = dontSplit

        # beam definition
        self.particle = particle
        self.beamdefined = False
        self.correctedbeamdef = False

        # File input and output
        self.file = inputfile
        self.fileloaded = False
        self.gmadoutput = gmad
        self.gmadDir = gmadDir
        self.madxoutput = madx
        self.madxDir = madxDir
        self.numberparts = -1

        # transport optics output is modified to be a single line
        self.singleLineOptics = False
        self.keepName = keepName
        self.combineDrifts = combineDrifts


class _Registry:
    """
    A class used as a registry during conversion.
    """
    def __init__(self):
        self.elements = []
        self.names = []
        self.lines = []
        self.length = []
        self._uniquenames = []
        self._totalLength = 0

    def AddToRegistry(self, linedict, line):
        if not isinstance(linedict, dict):
            raise TypeError("Added element is not a Dictionary")
        self.elements.append(linedict)
        self.names.append(linedict['name'])
        if not linedict['name'] in self._uniquenames:
            self._uniquenames.append(linedict['name'])

        self.lines.append(line)
        # Cumulative length
        length = round(linedict['length'], 5)
        if len(self.length) > 0:
            self.length.append(length + self._totalLength)
        else:
            self.length.append(length)
        self._totalLength += length

    def GetElementIndex(self, name):
        elenums = []
        if name not in self.names:
            return elenums
        else:
            # Add all elements of the same name as a single element may be
            # used multiple times.
            for index, elename in enumerate(self.names):
                if elename == name:
                    elenums.append(index)
            return elenums

    def GetElement(self, name):
        elenum = self.GetElementIndex(name)
        if isinstance(elenum, list):
            elementList = []
            for num in elenum:
                elementList.append(self.elements[num])
            return elementList
        else:
            return self.elements[elenum]

    def GetElementEndSPosition(self, name):
        elenum = self.GetElementIndex(name)
        if isinstance(elenum, list):
            elementList = []
            for num in elenum:
                elementList.append(self.length[num])
            return elementList
        else:
            return self.length[elenum]

    def GetElementStartSPosition(self, name):
        elenum = self.GetElementIndex(name)
        endS = self.GetElementEndSPosition(name)

        if isinstance(elenum, list):
            elementList = []
            for index, num in enumerate(elenum):
                element = self.elements[num]
                length = element['length']
                startS = endS[index] - length
                elementList.append(round(startS, 5))
            return elementList
        else:
            element = self.elements[elenum]
            length = element['length']
            startS = endS - length
            return round(startS, 5)

    def UpdateLength(self, linedict):
        """
        Function to increases the machines length, but does not add element data.
        This is so the S positions of named elements in the fitting registry can
        be calculated correctly.
        """
        if not isinstance(linedict, dict):
            raise TypeError("Added element is not a Dictionary")
        self._totalLength += linedict['length']



"""
pytransport - Royal Holloway utility to manipulate TRANSPORT data and models.

Authors:

 * William Shields
 * Jochem Snuverink

Copyright Royal Holloway, University of London 2018.

"""

__version__ = "1.1"

import _General
import Compare
import Convert
import Data
import Reader

__all__ = ['Compare',
           'Convert',
           'Data',
           'Reader']



from setuptools import setup, find_packages

setup(
    name='pytransport',
    version='1.1',
    packages=find_packages(exclude=["docs", "tests", "obsolete"]),
    # Not sure how strict these need to be...
    install_requires=["matplotlib",
                      "numpy",
                      "scipy"],
    # Some version of python2.7
    python_requires="==2.7.*",

    author='JAI@RHUL',
    author_email='william.shields.2010@live.rhul.ac.uk',
    description="Convert TRANSPORT models and load TRANSPORT output.",
    license='GPL3',
    url='https://bitbucket.org/jairhul/pytransport/'
)



# pytransport.Elements - tools for Transport element conversion
# Version 1.0
# W. Shields and J. Snuverink
# william.shields.2010@live.rhul.ac.uk

"""
Wrapper function for automatically converting to bdsim and/or madx

Class containing functions to help convert elements from Transport to gmad/madx.
Class instantiated in pybdsim.Convert.Transport2Gmad and pymadx.Convert.Transport2Madx.

Classes:
_Convert - a class used to convert different element types.

"""

import numpy as _np
import string as _str

import _General
from _General import _Writer
from Data import ConversionData as _convData
from pytransport import Reader as _Reader


def Convert(inputfile,
            particle='proton',
            distrType='gauss',
            output='bdsim',
            outputDir='',
            debug=False,
            dontSplit=False,
            keepName=False,
            combineDrifts=False,
            options=None,
            machine=None):
    """
    **Convert** convert a Transport input or output file into an input file for bdsim, madx, or both

    +-------------------------------+-------------------------------------------------------------------+
    | **inputfile**                 | dtype = string                                                    |
    |                               | path to the input file                                            |
    +-------------------------------+-------------------------------------------------------------------+
    | **particle**                  | dtype = string. Optional, default = "proton"                      |
    |                               | the particle species                                              |
    +-------------------------------+-------------------------------------------------------------------+
    | **distrType**                 | dtype = string. Optional, Default = "gauss".                      |
    |                               | the beam distribution type. Can be either gauss or gausstwiss.    |
    +-------------------------------+-------------------------------------------------------------------+
    | **output**                    | dtype=string. Optional, default = "bdsim"                         |
    |                               | the output type, can be "bdsim", "madx", or "both"                |
    +-------------------------------+-------------------------------------------------------------------+
    | **outputDir**                 | dtype=string. Optional, default = ""                              |
    |                               | the output directory where the files will be written              |
    |                               | if no input supplied, defaults to output type. if outputDir is    |
    |                               | supplied and output is "both", outputDir is appended with the     |
    |                               | output type, e.g. outputDir_bdsim                                 |
    +-------------------------------+-------------------------------------------------------------------+
    | **debug**                     | dtype = bool. Optional, default = False                           |
    |                               | output a log file (inputfile_conversion.log) detailing the        |
    |                               | conversion process, element by element                            |
    +-------------------------------+-------------------------------------------------------------------+
    | **dontSplit**                 | dtype = bool. Optional, default = False                           |
    |                               | the converter splits the machine into multiple parts when a beam  |
    |                               | is redefined in a Transport lattice. dontSplit overrides this and |
    |                               | forces the machine to be written to a single file                 |
    +-------------------------------+-------------------------------------------------------------------+
    | **keepName**                  | dtype = bool. Optional, default = False                           |
    |                               | keep the names of elements as defined in the Transport inputfile. |
    |                               | Appends element name with _N where N is an integer if the element |
    |                               | name has already been used                                        |
    +-------------------------------+-------------------------------------------------------------------+
    | **combineDrifts**             | dtype = bool. Optional, default = False                           |
    |                               | combine multiple consecutive drifts into a single drift           |
    +-------------------------------+-------------------------------------------------------------------+
    | **machine**                   | dtype = pybdsim.Builder.Machine or pymadx.Builder.Machine         |
    |                               | required, default = None                                          |
    |                               | machine instance required for conversion.                         |
    +-------------------------------+-------------------------------------------------------------------+
    | **options**                   | dtype = pybdsim.Options.Options. Optional, default = None         |
    |                               | options instance required to write options in bdsim conversion.   |
    |                               | Ignored if converting to "madx" format.                           |
    +-------------------------------+-------------------------------------------------------------------+

    Example:

    >>> Convert(inputfile, machine=pybdsim.Builder.Machine(), options=pybdsim.Builder.Options())
    >>> Convert(inputfile, output="madx", machine=pymadx.Builder.Machine())

    Writes converted machine to disk. Reader automatically detects if the supplied input file is a Transport input
    file or Transport output file.

    """
    outputType = _str.lower(output)

    if machine is None:
        raise TypeError("machine instance must be supplied")
    if ((outputType == 'bdsim') or (output == 'both')) and (options is None):
        raise TypeError("pybdsim.Options.Options must be supplied for bdsim conversion")
    if (outputType == 'madx') and (options is not None):
        print("Ignoring supplied options for madx conversion")

    # default for output='bdsim'
    gmadDir = outputDir
    madxDir = outputDir

    if outputType == 'bdsim':
        gmad = True
        madx = False
        if outputDir == '':
            gmadDir = 'bdsim'
            madxDir = ''
    elif outputType == 'madx':
        gmad = False
        madx = True
        if outputDir == '':
            gmadDir = ''
            madxDir = 'madx'
    elif outputType == 'both':
        gmad = True
        madx = True
        if outputDir == '':
            gmadDir = 'bdsim'
            madxDir = 'madx'
        else:
            gmadDir += '_bdsim'
            madxDir += '_madx'
    else:
        raise IOError("Unknown output type '"+output+"'")

    converter = _Convert(_convData(inputfile=inputfile,
                                   particle=particle,
                                   distrType=distrType,
                                   gmad=gmad,
                                   madx=madx,
                                   gmadDir=gmadDir,
                                   madxDir=madxDir,
                                   debug=debug,
                                   dontSplit=dontSplit,
                                   keepName=keepName,
                                   combineDrifts=combineDrifts,
                                   options=options,
                                   machine=machine))
    converter.Convert()


class _Convert:
    """
    Class for converting different Transport element types into gmad/madx format.

    Required: transportData, dtype = pytransport.Data.ConversionData

    """
    def __init__(self, transportData):
        if not isinstance(transportData, _convData):
            raise TypeError("transportData must be a pytransport.Data.ConversionData instance")
        self.Transport = transportData
        logfileName = _General.RemoveFileExt(self.Transport.convprops.file) + '_conversion.log'
        self.Writer = _Writer(debugOutput=self.Transport.convprops.debug,
                              writeToLog=self.Transport.convprops.outlog,
                              logfile=logfileName)

    def LoadFile(self, inputfile):
        """
        Load a Transport file.
        """
        if not isinstance(inputfile, _np.str):
            raise TypeError("Input must be a string")

        infile = inputfile.split('/')[-1]  # Remove filepath, leave just filename
        self.Transport._file = infile[:-4]  # Remove extension
        self.Transport._filename = inputfile
        isOutput = _General.CheckIsOutput(inputfile)  # Is a TRANSPORT standard output file.
        self.Writer.DebugPrintout("File Read.")
        if isOutput:
            lattice = _Reader.GetLattice(inputfile)
            fitres = _Reader.GetResultsFromFitting(inputfile)
            self.Transport = _General.OutputFitsToRegistry(self.Transport, fitres)
            self.Writer.DebugPrintout('Adding any fitting output to the fitting registry (self.FitRegistry)')

            self.Writer.DebugPrintout('Processing file and adding to Transport class.')
            for linenum, latticeline in enumerate(lattice):
                latticeline = latticeline.replace(';', '')
                line = _np.array(latticeline.split(' '), dtype=_np.str)
                line = _General.RemoveIllegals(line)

                # Method of dealing with split lines in the output
                # Should only be applicable to type 12 entry (up to 15 variables)
                # It is assumed that the line is always split, so be careful.
                prevline = lattice[linenum - 1].replace(';', '')
                prevline = _np.array(prevline.split(' '), dtype=_np.str)
                prevline = _General.RemoveIllegals(prevline)

                try:
                    if (linenum > 0) and _np.abs(_np.float(line[0])) == 12.0:
                        latticeline, line = _General.JoinSplitLines(linenum, lattice)
                    # Ignore line after type 12 entry (second part of split line)
                    if (linenum > 1) and _np.abs(_np.float(prevline[0])) == 12.0:
                        pass
                    else:
                        self.Transport.data.append(line)
                        self.Transport.filedata.append(latticeline)
                except ValueError:
                    self.Transport.data.append(line)
                    self.Transport.filedata.append(latticeline)
                except IndexError:
                    pass

        else:
            f = open(inputfile)
            for inputline in f:
                endoflinepos = _General.FindEndOfLine(inputline)
                templine = inputline
                if endoflinepos > 0:
                    templine = inputline[:endoflinepos]
                line = _np.array(templine.split(' '), dtype=_np.str)
                # do not change comment lines
                if not line[0][0] == '(':
                    line = _General.RemoveIllegals(line)
                self.Transport.data.append(line)
                self.Transport.filedata.append(inputline)
            f.close()
        self.Transport.convprops.fileloaded = True

    def Write(self):
        """
        Write the converted TRANSPORT file to disk.
        """
        self.Writer.DebugPrintout("Adding beam to gmad machine:")
        self.Transport.AddBeam()
        if self.Transport.convprops.gmadoutput:
            self.Transport.AddOptions()
        self.Transport.machine.AddSampler('all')
        self.Writer.BeamDebugPrintout(self.Transport.beamprops, self.Transport.units)
        fname = _General.RemoveFileExt(self.Transport.convprops.file)
        if self.Transport.convprops.numberparts < 0:
            filename = fname
        else:
            self.Transport.convprops.numberparts += 1
            filename = fname + '_part' + _np.str(self.Transport.convprops.numberparts)
        self.Writer.Write(self.Transport, filename)

    def Convert(self):
        """
        Convert, process, and write.
        """
        self.LoadFile(self.Transport.convprops.file)

        if not self.Transport.convprops.fileloaded:
            self.Writer.Printout('No file loaded.')
            return
        self.ProcessAndBuild()
        self.Write()

    def ProcessAndBuild(self):
        """
        Function that loops over the lattice, adds the elements to the element registry,
        and updates any elements that have fitted parameters.
        It then converts the registry elements and adds to the gmad machine.
        """
        self.Writer.DebugPrintout('Processing tokenised lines from input file and adding to element registry.\n')

        for linenum, line in enumerate(self.Transport.data):
            self.Writer.DebugPrintout('Processing tokenised line ' + _np.str(linenum) + ' :')
            self.Writer.DebugPrintout('\t' + str(line))
            self.Writer.DebugPrintout('\tOriginal :')
            self.Writer.DebugPrintout('\t' + self.Transport.filedata[linenum])

            # Checks if the SENTINEL line is found. SENTINEL relates to TRANSPORT fitting routine and is only written
            # after the lattice definition, so there's no point reading lines beyond it.
            if _General.CheckIsSentinel(line):
                self.Writer.DebugPrintout('Sentinel Found.')
                break
            # Test for positive element, negative ones ignored in TRANSPORT so ignored here too.
            try:
                typeNum = _General.GetTypeNum(line)
                if typeNum > 0:
                    if self.Transport.data[0][0] == 'OUTPUT':
                        self._ElementPrepper(line, linenum, 'output')
                    else:
                        line = _General.RemoveIllegals(line)
                        self._ElementPrepper(line, linenum, 'input')
                else:
                    self.Writer.DebugPrintout('\tType code is 0 or negative, ignoring line.')
            except ValueError:
                errorline = '\tCannot process line ' + _np.str(linenum) + ', '
                if line[0][0] == '(' or line[0][0] == '/':
                    errorline += 'line is a comment.'
                elif line[0][0] == 'S':  # S used as first character in SENTINEL command.
                    errorline = 'line is for TRANSPORT fitting routine.'
                elif line[0] == '\n':
                    errorline = 'line is blank.'
                else:
                    errorline = 'reason unknown.'
                self.Writer.DebugPrintout(errorline)
        self._UpdateElementsFromFits()
        self.Writer.DebugPrintout(
            'Converting registry elements to pybdsim compatible format and adding to machine builder.\n')

        skipNextDrift = False  # used for collimators
        lastElementWasADrift = True  # default value
        if self.Transport.convprops.combineDrifts:
            lastElementWasADrift = False
        for linenum, linedict in enumerate(self.Transport.ElementRegistry.elements):
            self.Writer.DebugPrintout('Converting element number ' + _np.str(linenum) + ':')
            convertline = '\t'
            for keynum, key in enumerate(linedict.keys()):
                if keynum != 0:
                    convertline += ', '
                if key == 'data':
                    convertline += 'element data:'
                    for ele in linedict[key]:
                        convertline += ('\t' + _np.str(ele))
                else:
                    convertline += (key + ': ' + _np.str(linedict[key]))
                if keynum == len(linedict.keys()):
                    convertline += '.'
            self.Writer.DebugPrintout(convertline)

            if self.Transport.convprops.combineDrifts:
                if lastElementWasADrift and linedict['elementnum'] != 3.0 and linedict['elementnum'] < 20.0:
                    # write possibly combined drift
                    self.Writer.DebugPrintout('\n\tConvert delayed drift(s)')
                    self.Drift(linedictDrift)
                    lastElementWasADrift = False
                    self.Writer.DebugPrintout('\n\tNow convert element number' + _np.str(linenum))

            if linedict['elementnum'] == 15.0:
                self.UnitChange(linedict)
            if linedict['elementnum'] == 20.0:
                self.ChangeBend(linedict)
            if linedict['elementnum'] == 1.0:  # Add beam on first definition
                if not self.Transport.convprops.beamdefined:
                    self.DefineBeam(linedict)
                elif not self.Transport.convprops.dontSplit:  # Only update beyond first definition if splitting is permitted
                    self.DefineBeam(linedict)
            if linedict['elementnum'] == 3.0:
                if skipNextDrift:
                    skipNextDrift = False
                    continue
                if self.Transport.convprops.combineDrifts:
                    self.Writer.DebugPrintout('\tDelay drift')
                    if lastElementWasADrift:
                        linedictDrift['length'] += linedict['length']  # update linedictDrift
                        if not linedictDrift['name']:
                            linedictDrift['name'] = linedict['name']  # keep first non-empty name
                    else:
                        linedictDrift = linedict  # new linedictDrift
                        lastElementWasADrift = True
                else:
                    self.Drift(linedict)
            if linedict['elementnum'] == 4.0:
                self.Dipole(linedict)
            if linedict['elementnum'] == 5.0:
                self.Quadrupole(linedict)
            if linedict['elementnum'] == 6.0:
                if not self.Transport.convprops.typeCode6IsTransUpdate:
                    self.Collimator(linedict)
                    # Length gotten from next drift
                    if linedict['length'] > 0.0:
                        skipNextDrift = True
                else:
                    self.TransformUpdate(linedict)
            if linedict['elementnum'] == 12.0:
                self.Correction(linedict)
            if linedict['elementnum'] == 11.0:
                self.Acceleration(linedict)
            if linedict['elementnum'] == 13.0:
                self.Printline(linedict)
            if linedict['elementnum'] == 16.0:
                self.SpecialInput(linedict)
            if linedict['elementnum'] == 18.0:
                self.Sextupole(linedict)
            if linedict['elementnum'] == 19.0:
                self.Solenoid(linedict)

            # 9.  : 'Repetition' - for nesting elements
            if linedict['elementnum'] == 9.0:
                self.Writer.DebugPrintout('\tWARNING Repetition Element not implemented in converter!')
            if linedict['elementnum'] == 2.0:
                errorline = '\tLine is a poleface rotation which is handled by the previous or next dipole element.'
                self.Writer.DebugPrintout(errorline)

            self.Writer.DebugPrintout('\n')

        # OTHER TYPES WHICH CAN BE IGNORED:
        # 6.0.X : Update RX matrix used in TRANSPORT
        # 7.  : 'Shift beam centroid'
        # 8.  : Magnet alignment tolerances
        # 10. : Fitting constraint
        # 14. : Arbitrary transformation of TRANSPORT matrix
        # 22. : Space charge element
        # 23. : RF Cavity (Buncher), changes bunch energy spread

        # Write also last drift
        if self.Transport.convprops.combineDrifts:
            if lastElementWasADrift:
                self.Writer.DebugPrintout('\tConvert delayed drift(s)')
                self.Drift(linedictDrift)

    def _ElementPrepper(self, line, linenum, filetype='input'):
        """
        Function to extract the data and prepare it for processing by each element function.
        This has been written as the lattice lines from an input file and output file are different,
        so it just a way of correctly ordering the information.
        """
        linedict = {'elementnum': 0.0,
                    'name': '',
                    'length': 0.0,
                    'isZeroLength': True}
        numElements = _np.str(len(self.Transport.ElementRegistry.elements))
        typeNum = _General.GetTypeNum(line)
        linedict['elementnum'] = typeNum

        if typeNum == 15.0:
            label = _General.GetLabel(line)
            if filetype == 'output':
                linedict['label'] = line[2].strip('"')
            if filetype == 'input':
                linedict['label'] = label
            linedict['number'] = line[1]
            self.Writer.ElementPrepDebugPrintout("Unit Control", numElements)

        if typeNum == 20.0:
            angle = 0  # Default
            if len(line) == 2:  # i.e has a label
                endofline = _General.FindEndOfLine(line[1])
                angle = line[1][:endofline]
            else:
                for index in line[1:]:
                    try:
                        angle = _np.float(index)
                        break
                    except ValueError:
                        pass
            linedict['angle'] = angle
            self.Writer.ElementPrepDebugPrintout("coordinate rotation", numElements)

        if typeNum == 1.0:
            linedict['name'] = _General.GetLabel(line)
            linedict['isAddition'] = False
            if _General.CheckIsAddition(line, filetype):
                linedict['isAddition'] = True
            # line = self._remove_label(line)
            if len(line) < 8:
                raise IndexError("Incorrect number of beam parameters.")
            n = 1
            if filetype == 'input':
                n = 0
            elif filetype == 'output':
                n = 1

            linedict['momentum'] = line[7 + n]
            linedict['Sigmax'] = line[1 + n]
            linedict['Sigmay'] = line[3 + n]
            linedict['Sigmaxp'] = line[2 + n]
            linedict['Sigmayp'] = line[4 + n]
            linedict['SigmaT'] = line[5 + n]
            linedict['SigmaE'] = line[6 + n]
            self.Writer.ElementPrepDebugPrintout("Beam definition or r.m.s addition", numElements)

        if typeNum == 2.0:
            linedict['name'] = _General.GetLabel(line)
            linedict['data'] = _General.GetElementData(line)
            self.Writer.ElementPrepDebugPrintout("poleface rotation", numElements)

        if typeNum == 3.0:
            linedict['name'] = _General.GetLabel(line)
            data = _General.GetElementData(line)
            linedict['length'] = data[0]
            linedict['isZeroLength'] = False
            self.Writer.ElementPrepDebugPrintout("drift", numElements)

        if typeNum == 4.0:
            linedict['name'] = _General.GetLabel(line)
            linedict['linenum'] = linenum
            data = _General.GetElementData(line)
            linedict['data'] = data
            linedict['length'] = data[0]
            linedict['isZeroLength'] = False
            e1, e2 = _General.GetFaceRotationAngles(self.Transport.data, linenum)
            linedict['e1'] = e1
            linedict['e2'] = e2
            self.Writer.ElementPrepDebugPrintout("dipole", numElements)

        if typeNum == 5.0:
            linedict['name'] = _General.GetLabel(line)
            data = _General.GetElementData(line)
            linedict['data'] = data
            linedict['length'] = data[0]
            linedict['isZeroLength'] = False
            self.Writer.ElementPrepDebugPrintout("quadrupole", numElements)

        if typeNum == 6.0:
            # element is a collimator or transform update
            # transform update is later ignored so only update linedict as if collimator

            physicalElements = [1.0, 3.0, 4.0, 5.0, 11.0, 18.0, 19.0]

            # Only iterate if not the last element
            if linenum == len(self.Transport.data):
                pass
            else:
                # Since collimators have zero length in TRANSPORT, chosen to use length of next drift instead if
                # present. Check all remaining elements for the next drift, following element(s) may be non-physical
                # element which can be ignored as it shouldnt affect the beamline. Ignore beam definition too in
                # the case where machine splitting is not permitted.
                for nextline in self.Transport.data[linenum + 1:]:
                    nextTypeNum = _General.GetTypeNum(nextline)
                    if nextTypeNum == 3.0:
                        nextData = _General.GetElementData(nextline)
                        linedict['length'] = nextData[0]
                        linedict['isZeroLength'] = False
                        linedict['name'] = _General.GetLabel(line)
                        data = _General.GetElementData(line)
                        linedict['data'] = data
                        break
                    # stop if physical element or beam redef if splitting permitted
                    elif nextTypeNum in physicalElements:
                        if (nextTypeNum == 1.0) and self.Transport.convprops.dontSplit:
                            pass
                        elif (nextTypeNum == 6.0) and self.Transport.convprops.typeCode6IsTransUpdate:
                            pass
                        else:
                            break
                    # ignore non-physical element
                    else:
                        pass
            # Can be either transform update or collimator, a 16. 14 entry changes the definition but is only
            # processed after ALL elements are added to the registry.
            self.Writer.ElementPrepDebugPrintout("transform update or collimator", numElements)

        if typeNum == 9.0:
            self.Writer.ElementPrepDebugPrintout("repetition control", numElements)

        if typeNum == 11.0:
            linedict['name'] = _General.GetLabel(line)
            data = _General.GetElementData(line)
            linedict['data'] = data
            linedict['length'] = data[0]
            linedict['voltage'] = data[1]
            linedict['isZeroLength'] = False
            if len(data) == 4:  # Older case for single element
                linedict['phase_lag'] = data[2]
                linedict['wavel'] = data[3]
            self.Writer.ElementPrepDebugPrintout("acceleration element", numElements)

        if typeNum == 12.0:
            linedict['data'] = _General.GetElementData(line)
            linedict['name'] = _General.GetLabel(line)

            prevline = self.Transport.data[linenum - 1]  # .split(' ')
            linedict['prevlinenum'] = _np.float(prevline[0])
            linedict['isAddition'] = False
            if _General.CheckIsAddition(line):
                linedict['isAddition'] = True
            self.Writer.ElementPrepDebugPrintout("beam rotation", numElements)

        if typeNum == 13.0:
            linedict['data'] = _General.GetElementData(line)
            self.Writer.ElementPrepDebugPrintout("Input/Output control", numElements)

        if typeNum == 16.0:
            linedict['data'] = _General.GetElementData(line)
            self.Writer.ElementPrepDebugPrintout("special input", numElements)

        if typeNum == 18.0:
            linedict['name'] = _General.GetLabel(line)
            data = _General.GetElementData(line)
            linedict['data'] = data
            linedict['length'] = data[0]
            linedict['isZeroLength'] = False
            self.Writer.ElementPrepDebugPrintout("sextupole", numElements)

        if typeNum == 19.0:
            linedict['name'] = _General.GetLabel(line)
            data = _General.GetElementData(line)
            linedict['data'] = data
            linedict['length'] = data[0]
            linedict['isZeroLength'] = False
            self.Writer.ElementPrepDebugPrintout("solenoid", numElements)

        if typeNum == 22.0:
            self.Writer.ElementPrepDebugPrintout("space charge element", numElements)

        if typeNum == 23.0:
            self.Writer.ElementPrepDebugPrintout("buncher", numElements)

        rawline = self.Transport.filedata[linenum]
        self.Transport.ElementRegistry.AddToRegistry(linedict, rawline)

    def DefineBeam(self, linedict):
        if linedict['isAddition']:
            if self.Transport.convprops.debug:
                self.Writer.DebugPrintout('\tIgnoring beam rms addition.')
            return
        if self.Transport.convprops.beamdefined and not self.Transport.convprops.dontSplit:
            self.Writer.Printout('Beam redefinition found. Writing previous section to file.')
            self.Writer.Printout('Splitting into multiple machines.')
            self.Transport.convprops.numberparts += 1
            self.Transport.AddBeam()
            if self.Transport.convprops.gmadoutput:
                self.Transport.AddOptions()
            self.Transport.machine.AddSampler('all')
            self.Writer.BeamDebugPrintout(self.Transport.beamprops, self.Transport.units)
            fname = _General.RemoveFileExt(self.Transport.convprops.file)
            if self.Transport.convprops.numberparts < 0:
                filename = fname
            else:
                self.Transport.convprops.numberparts += 1
                filename = fname + '_part' + _np.str(self.Transport.convprops.numberparts)
            self.Writer.Write(self.Transport, filename)
            self.Transport.ResetMachine()
            self.Transport.convprops.correctedbeamdef = False

        momentum = linedict['momentum']

        self.Transport.convprops.beamdefined = True

        # Convert momentum to energy and set distribution params.
        self.Transport = _General.UpdateEnergyFromMomentum(self.Transport, momentum)
        self.Transport.beamprops.SigmaX  = _np.float(linedict['Sigmax'])
        self.Transport.beamprops.SigmaY  = _np.float(linedict['Sigmay'])
        self.Transport.beamprops.SigmaXP = _np.float(linedict['Sigmaxp'])
        self.Transport.beamprops.SigmaYP = _np.float(linedict['Sigmayp'])
        self.Transport.beamprops.SigmaE  = _np.float(linedict['SigmaE']) * 0.01 * (self.Transport.beamprops.beta**2)  # Convert from percentage mom spread to absolute espread
        self.Transport.beamprops.SigmaT  = _General.ConvertBunchLength(self.Transport, _np.float(linedict['SigmaT']))  # Get bunch length in seconds.

        # Calculate Initial Twiss params
        try:
            self.Transport.beamprops.betx = self.Transport.beamprops.SigmaX / self.Transport.beamprops.SigmaXP
        except ZeroDivisionError:
            self.Transport.beamprops.betx = 0
        try:
            self.Transport.beamprops.bety = self.Transport.beamprops.SigmaY / self.Transport.beamprops.SigmaYP
        except ZeroDivisionError:
            self.Transport.beamprops.bety = 0
        self.Transport.beamprops.emitx = self.Transport.beamprops.SigmaX * self.Transport.beamprops.SigmaXP / 1000.0
        self.Transport.beamprops.emity = self.Transport.beamprops.SigmaY * self.Transport.beamprops.SigmaYP / 1000.0

        self.Writer.BeamDebugPrintout(self.Transport.beamprops, self.Transport.units)

    def Drift(self, linedict):
        driftlen = linedict['length']
        if driftlen <= 0:
            self.Writer.DebugPrintout('\tZero or negative length element, ignoring.')
            return

        lenInM = driftlen * _General.ScaleToMeters(self.Transport, 'element_length')  # length in metres

        self.Transport.machineprops.drifts += 1
        elementid = self._GetTransportElementName(linedict['name'])
        if not elementid:  # check on empty string
            elementid = 'DR'+_np.str(self.Transport.machineprops.drifts)

        # pybdsim and pymadx are the same.
        self.Transport.machine.AddDrift(name=elementid, length=lenInM)

        self.Writer.DebugPrintout('\tConverted to:')
        self.Writer.DebugPrintout('\t' + 'Drift ' + elementid + ', length ' + _np.str(lenInM) + ' m')

    def Dipole(self, linedict):
        linenum = linedict['linenum']
        dipoledata = linedict['data']
        length = dipoledata[0]  # First two non-blanks must be the entries in a specific order.

        # Get poleface rotation
        e1 = linedict['e1'] * ((_np.pi / 180.0)*self.Transport.machineprops.bending)  # Entrance pole face rotation.
        e2 = linedict['e2'] * ((_np.pi / 180.0)*self.Transport.machineprops.bending)  # Exit pole face rotation.

        if e1 != 0:
            self.Writer.DebugPrintout('\tPreceding element on line (' + _np.str(linenum-1) + ') of the inout file provides an entrance poleface rotation of ' + _np.str(_np.round(e1, 4)) + ' rad.')
        if e2 != 0:
            self.Writer.DebugPrintout('\tFollowing element on line (' + _np.str(linenum+1) + ') of the input file provides an exit poleface rotation of ' + _np.str(_np.round(e2, 4)) + ' rad.')

        # Fringe Field Integrals
        fintVal = 0
        fintxVal = 0
        fintSecVal = 0
        fintxSecVal = 0
        if e1 != 0:
            fintVal = self.Transport.machineprops.fringeIntegral
            fintSecVal = self.Transport.machineprops.secondfringeInt
        if e2 != 0:
            fintxVal = self.Transport.machineprops.fringeIntegral
            fintxSecVal = self.Transport.machineprops.secondfringeInt

        if (fintVal != 0) or (fintxVal != 0):
            self.Writer.DebugPrintout('\tA previous entry set the fringe field integral K1=' + _np.str(self.Transport.machineprops.fringeIntegral) + '.')
            self.Writer.DebugPrintout('\tA previous entry set the second fringe field integral K2=' + _np.str(self.Transport.machineprops.secondfringeInt) + '.')

        hgap = self.Transport.machineprops.dipoleVertAper * self.Transport.scale[self.Transport.units['bend_vert_gap'][0]]

        # Poleface curvatures
        h1 = self.Transport.machineprops.bendInCurvature
        h2 = self.Transport.machineprops.bendOutCurvature
        if self.Transport.machineprops.bendInCurvature != 0:
            self.Writer.DebugPrintout('\tA previous entry set the dipole poleface entrance curvature H1=' + _np.str(self.Transport.machineprops.bendInCurvature) + '.')
        if self.Transport.machineprops.bendOutCurvature != 0:
            self.Writer.DebugPrintout('\tA previous entry set the dipole poleface entrance curvature H2=' + _np.str(self.Transport.machineprops.bendOutCurvature) + '.')

        # Calculate bending angle
        if self.Transport.machineprops.benddef:
            bfield = dipoledata[1]
            field_in_Gauss = bfield * self.Transport.scale[self.Transport.units['magnetic_fields'][0]]  # Scale to Gauss
            field_in_Tesla = field_in_Gauss * 1e-4                                  # Convert to Tesla
            if field_in_Tesla == 0:
                angle = 0                                                           # zero field = zero angle
            else:
                rho = self.Transport.beamprops.brho / (_np.float(field_in_Tesla))         # Calculate bending radius.
                angle = (_np.float(length) / rho) * self.Transport.machineprops.bending   # for direction of bend
            self.Writer.DebugPrintout('\tbfield = ' + _np.str(field_in_Gauss) + ' kG')
            self.Writer.DebugPrintout('\tbfield = ' + _np.str(field_in_Tesla) + ' T')
            self.Writer.DebugPrintout('\tCorresponds to angle of ' + _np.str(_np.round(angle, 4)) + ' rad.')
        else:
            angle_in_deg = dipoledata[1]
            angle = angle_in_deg * (_np.pi/180.) * self.Transport.machineprops.bending

        # Convert element length
        lenInM = length * _General.ScaleToMeters(self.Transport, 'element_length')

        self.Transport.machineprops.dipoles += 1
        elementid = self._GetTransportElementName(linedict['name'])
        if not elementid:  # check on empty string
            elementid = 'BM' + _np.str(self.Transport.machineprops.dipoles)

        # pybdsim and pymadx set differently depending on second fringe field integral. Check for non zero pole face rotation.
        if (e1 != 0) and (e2 != 0):
            if self.Transport.convprops.madxoutput:
                self.Transport.machine.AddDipole(name=elementid, category='sbend', length=lenInM,
                                                 angle=_np.round(angle, 4), e1=_np.round(e1, 4), e2=_np.round(e2, 4),
                                                 fint=fintVal, fintx=fintxVal, hgap=hgap)
            else:
                self.Transport.machine.AddDipole(name=elementid, category='sbend', length=lenInM,
                                                 angle=_np.round(angle, 4), e1=_np.round(e1, 4), e2=_np.round(e2, 4),
                                                 fint=fintVal, fintx=fintxVal, fintK2=fintSecVal, fintxK2=fintxSecVal,
                                                 hgap=hgap)
        elif (e1 != 0) and (e2 == 0):
            if self.Transport.convprops.madxoutput:
                self.Transport.machine.AddDipole(name=elementid, category='sbend', length=lenInM,
                                                 angle=_np.round(angle, 4), e1=_np.round(e1, 4), fint=fintVal, fintx=0,
                                                 hgap=hgap)
            else:
                self.Transport.machine.AddDipole(name=elementid, category='sbend', length=lenInM,
                                                 angle=_np.round(angle, 4), e1=_np.round(e1, 4), fint=fintVal, fintx=0,
                                                 fintK2=fintSecVal, fintxK2=0, hgap=hgap)
        elif (e1 == 0) and (e2 != 0):
            if self.Transport.convprops.madxoutput:
                self.Transport.machine.AddDipole(name=elementid, category='sbend', length=lenInM,
                                                 angle=_np.round(angle, 4), e2=_np.round(e2, 4), fint=0, fintx=fintxVal,
                                                 hgap=hgap)
            else:
                self.Transport.machine.AddDipole(name=elementid, category='sbend', length=lenInM,
                                                 angle=_np.round(angle, 4), e2=_np.round(e2, 4), fint=0, fintx=fintxVal,
                                                 fintK2=0, fintxK2=fintxSecVal, hgap=hgap)
        else:
            self.Transport.machine.AddDipole(name=elementid, category='sbend', length=lenInM, angle=_np.round(angle, 4))

        # Debug output
        if (e1 != 0) and (e2 != 0):
            polefacestr = ', e1= ' + _np.str(_np.round(e1, 4)) + ' rad, e2= ' + _np.str(_np.round(e2, 4)) + ' rad'
        elif (e1 != 0) and (e2 == 0):
            polefacestr = ', e1= ' + _np.str(_np.round(e1, 4)) + ' rad'
        elif (e1 == 0) and (e2 != 0):
            polefacestr = ', e2= ' + _np.str(_np.round(e2, 4)) + ' rad'
        else:
            polefacestr = ''

        if (fintVal != 0) and (fintxVal != 0):
            fringestr = ' , fint= ' + _np.str(fintVal) + ', fintx= ' + _np.str(fintxVal)
        elif (fintVal != 0) and (fintxVal == 0):
            fringestr = ' , fint= ' + _np.str(fintVal)
        elif (fintVal == 0) and (fintxVal != 0):
            fringestr = ' , fintx= ' + _np.str(fintxVal)
        else:
            fringestr = ''

        if (fintSecVal != 0) and (fintxSecVal != 0):
            fringestr2 = ' , fintK2= ' + _np.str(fintSecVal) + ', fintxK2= ' + _np.str(fintxSecVal)
        elif (fintVal != 0) and (fintxSecVal == 0):
            fringestr2 = ' , fintK2= ' + _np.str(fintSecVal)
        elif (fintVal == 0) and (fintxSecVal != 0):
            fringestr2 = ' , fintxK2= ' + _np.str(fintxSecVal)
        else:
            fringestr2 = ''

        self.Writer.DebugPrintout('\tConverted to:')
        debugstring = 'Dipole ' + elementid + ', length= ' + _np.str(lenInM) + ' m, angle= ' + \
                      _np.str(_np.round(angle, 4)) + ' rad' + polefacestr + fringestr + fringestr2
        self.Writer.DebugPrintout('\t' + debugstring)

    def ChangeBend(self, linedict):
        """
        Function to change the direction of the dipole bend. Can be a direction other than horizontal (i.e != n*pi).
        """
        # NOT FULLY TESTED.
        angle = linedict['angle']
        rotation = False
        elementid = ''

        self.Transport.machineprops.angle = _np.float(angle)
        if self.Transport.machineprops.angle >= 360:
            self.Transport.machineprops.angle = _np.mod(self.Transport.machineprops.angle, 360)
        if self.Transport.machineprops.angle <= -360:
            self.Transport.machineprops.angle = _np.mod(self.Transport.machineprops.angle, -360)

        if self.Transport.machineprops.angle == 180 or self.Transport.machineprops.angle == -180:  # If 180 degrees, switch bending angle
            self.Transport.machineprops.bending *= -1

        elif self.Transport.machineprops.angle != 0:  # If not 180 degrees, use transform3d.
            # self.machineprops.angle *= -1
            # For conversion to correct direction. Eg in TRANSPORT -90 is upwards, in BDSIM, 90 is upwards.
            anginrad = self.Transport.machineprops.angle * (_np.pi / 180)
            self.Transport.machineprops.transforms += 1
            elementid = self._GetTransportElementName(linedict['name'])
            if not elementid:  # check on empty string
                elementid = 't' + _np.str(self.Transport.machineprops.transforms)

            # only call for gmad, warning for madx
            if self.Transport.convprops.gmadoutput:
                self.Transport.machine.AddTransform3D(name=elementid, psi=anginrad)
            elif self.Transport.convprops.madxoutput:
                self.Writer.DebugPrintout('\tWarning, MadX Builder does not have Transform 3D!')

            rotation = True

        if rotation:
            self.Writer.DebugPrintout('\tConverted to:')
            debugstring = '\tTransform3D ' + elementid + ', angle ' + _np.str(_np.round(self.Transport.machineprops.angle, 4)) + ' rad'
            self.Writer.DebugPrintout('\t'+debugstring)
        elif self.Transport.machineprops.angle == 180:
            self.Writer.DebugPrintout('\tBending direction set to Right')
        elif self.Transport.machineprops.angle == -180:
            self.Writer.DebugPrintout('\tBending direction set to Left')

    def Quadrupole(self, linedict):
        quaddata = linedict['data']
        length = quaddata[0]        # First three non-blanks must be the entries in a specific order.
        field_at_tip = quaddata[1]  # Field in TRANSPORT units
        pipe_rad = quaddata[2]      # Pipe Radius In TRANSPORT units

        field_in_Gauss = field_at_tip * self.Transport.scale[self.Transport.units['magnetic_fields'][0]]  # Convert to Gauss
        field_in_Tesla = field_in_Gauss * 1e-4  # Convert to Tesla

        pipe_in_metres = pipe_rad * _General.ScaleToMeters(self.Transport, 'bend_vert_gap')
        lenInM = length * _General.ScaleToMeters(self.Transport, 'element_length')

        field_gradient = (field_in_Tesla / pipe_in_metres) / self.Transport.beamprops.brho  # K1 in correct units

        self.Transport.machineprops.quads += 1

        elementid = self._GetTransportElementName(linedict['name'])
        if not elementid:  # check on empty string
            if field_gradient > 0:
                elementid = 'QF' + _np.str(self.Transport.machineprops.quads)
            elif field_gradient < 0:
                elementid = 'QD' + _np.str(self.Transport.machineprops.quads)
            else:
                elementid = 'NULLQUAD' + _np.str(self.Transport.machineprops.quads)  # For K1 = 0.

        # pybdsim and pymadx are the same.
        self.Transport.machine.AddQuadrupole(name=elementid, length=lenInM, k1=_np.round(field_gradient, 4))

        string1 = '\tQuadrupole, field in gauss = ' + _np.str(field_in_Gauss) + ' G, field in Tesla = ' + _np.str(field_in_Tesla) + ' T.'
        string2 = '\tBeampipe radius = ' + _np.str(pipe_in_metres) + ' m. Field gradient = '+ _np.str(field_in_Tesla/pipe_in_metres) + ' T/m.'
        string3 = '\tBrho = ' + _np.str(_np.round(self.Transport.beamprops.brho, 4)) + ' Tm. K1 = ' +_np.str(_np.round(field_gradient, 4)) + ' m^-2'
        self.Writer.DebugPrintout(string1)
        self.Writer.DebugPrintout(string2)
        self.Writer.DebugPrintout(string3)
        self.Writer.DebugPrintout('\tConverted to:')
        debugstring = 'Quadrupole ' + elementid + ', length= ' + _np.str(lenInM) + ' m, k1= ' + _np.str(_np.round(field_gradient, 4)) + ' T/m'
        self.Writer.DebugPrintout('\t' + debugstring)

    def Collimator(self, linedict):
        """
        A Function that writes the properties of a collimator element
        Only added for gmad, not for madx!
        """
        if linedict['length'] <= 0:
            self.Writer.DebugPrintout('\tZero or negative length element, ignoring.')
            return
        colldata = linedict['data']

        # Determine which entry is for horiz. and vert.
        aperx = self.Transport.machineprops.beampiperadius
        apery = self.Transport.machineprops.beampiperadius
        if _np.float(colldata[0]) == 1.0:
            aperx = colldata[1]
        elif _np.float(colldata[0]) == 3.0:
            apery = colldata[1]

        if len(colldata) > 2:
            if _np.float(colldata[2]) == 1.0:
                aperx = colldata[3]
            elif _np.float(colldata[2]) == 3.0:
                apery = colldata[3]
        aperx = _np.float(aperx)
        apery = _np.float(apery)

        lenInM = linedict['length'] * _General.ScaleToMeters(self.Transport, 'element_length')
        aperx_in_metres = aperx * _General.ScaleToMeters(self.Transport, 'x')
        apery_in_metres = apery * _General.ScaleToMeters(self.Transport, 'y')

        self.Transport.machineprops.collimators += 1
        elementid = self._GetTransportElementName(linedict['name'])
        if not elementid:  # check on empty string
            elementid = 'COL'+_np.str(self.Transport.machineprops.collimators)

        collimatorMaterial = 'copper'  # Default in BDSIM, added to prevent warnings
        # only call for gmad, warning for madx
        if self.Transport.convprops.gmadoutput:
            self.Transport.machine.AddRCol(name=elementid, length=lenInM, xsize=aperx_in_metres, ysize=apery_in_metres,
                                           material=collimatorMaterial)
        elif self.Transport.convprops.madxoutput:
            self.Writer.DebugPrintout('\tWarning, MadX Builder does not have RCOL')

        debugstring = '\tCollimator, x aperture = ' + _np.str(aperx_in_metres) \
                      + ' m, y aperture = ' + _np.str(apery_in_metres) + ' m.'
        self.Writer.DebugPrintout(debugstring)
        self.Writer.DebugPrintout('\tConverted to:')
        debugstring = 'Collimator ' + elementid + ', length= ' + _np.str(lenInM)\
                      + ' m, xsize= ' + _np.str(_np.round(aperx_in_metres, 4))
        debugstring += ' m, ysize= ' + _np.str(_np.round(apery_in_metres, 4)) + ' m.'
        self.Writer.DebugPrintout('\t' + debugstring)

    def Acceleration(self, linedict):
        """
        A Function that writes the properties of an acceleration element
        Only RF added for gmad, not for madx!
        """
        # Redundant function until comments and /or acceleration components can be handled

        accdata = linedict['data']
        acclen = linedict['length']
        e_gain = linedict['voltage']

        # TODO add phase_lag and wavelength to BDSIM

        # If zero length then start of a sequence, save total accelerating voltage
        if acclen == 0.0:
            self.Transport.convprops.isAccSequence = True
            self.Transport.machineprops._totalAccVoltage = e_gain
            self.Transport.machineprops._e_gain_prev = 0.0  # start at 0
            return

        if self.Transport.convprops.isAccSequence:
            # voltage means voltage relative to the end of this segment
            e_rel_gain = e_gain - self.Transport.machineprops._e_gain_prev
            self.Transport.machineprops._e_gain_prev = e_gain  # store for next segment
            if e_gain == 1.0:  # end of sequence
                self.Transport.convprops.isAccSequence = False

            e_gain = e_rel_gain * self.Transport.machineprops._totalAccVoltage

        gradient = e_gain * (self.Transport.scale[self.Transport.units['p_egain'][0]] / 1e6)
        gradient /= (acclen * self.Transport.scale[self.Transport.units['element_length'][0]])  # gradient in MV/m

        self.Transport.machineprops.rf += 1
        elname = "ACC" + _np.str(self.Transport.machineprops.rf)

        # only call for gmad, warning for madx
        if self.Transport.convprops.gmadoutput:
            self.Transport.machine.AddRFCavity(name=elname, length=acclen, gradient=gradient)
        elif self.Transport.convprops.madxoutput:
            self.Writer.DebugPrintout('\tWarning, MadX Builder does not have RF Cavity')

        # Update beam parameters
        self.Transport = _General.UpdateMomentumFromEnergy(self.Transport, self.Transport.beamprops.k_energy + e_gain)

        # Commented out untested code
        # if len(accdata) == 2:  # Newer case with multiple elements
        # self._acc_sequence(line)
        if len(accdata) == 4:     # Older case for single element
            phase_lag = linedict['phase_lag']
            wavel = linedict['wavel']

            # Write to file
            accline = '! An accelerator element goes here of length ' + \
                      _np.str(acclen) + ' ' + self.Transport.units['element_length'] + ', \n'
            accline2 = '! with an energy gain of ' + _np.str(e_gain) + ' ' + \
                       self.Transport.units['p_egain'] + ', phase lag of ' + _np.str(phase_lag) + ' degrees, \n'
            accline3 = '! and a wavelength of ' + _np.str(wavel) + ' ' + self.Transport.units['bunch_length'] + '. \n'

    def Sextupole(self, linedict):
        sextudata = linedict['data']
        length = sextudata[0]        # First three non-blanks must be the entries in a specific order.
        field_at_tip = sextudata[1]  # Field in TRANSPORT units
        pipe_rad = sextudata[2]      # Pipe Radius In TRANSPORT units

        field_in_Gauss = field_at_tip * self.Transport.scale[self.Transport.units['magnetic_fields'][0]]  # Convert to Gauss
        field_in_Tesla = field_in_Gauss * 1e-4  # Convert to Tesla

        pipe_in_metres = pipe_rad * _General.ScaleToMeters(self.Transport, 'bend_vert_gap')
        lenInM = length * _General.ScaleToMeters(self.Transport, 'element_length')

        field_gradient = (2*field_in_Tesla / pipe_in_metres**2) / self.Transport.beamprops.brho  # K2 in correct units

        self.Transport.machineprops.sextus += 1
        elementid = self._GetTransportElementName(linedict['name'])
        if not elementid:  # check on empty string
            elementid = 'SEXT'+_np.str(self.Transport.machineprops.sextus)

        # pybdsim and pymadx are the same.
        self.Transport.machine.AddSextupole(name=elementid, length=lenInM, k2=_np.round(field_gradient, 4))

        self.Writer.DebugPrintout('\tConverted to:')
        debugstring = 'Sextupole ' + elementid + ', length ' + _np.str(lenInM) + \
                      ' m, k2 ' + _np.str(_np.round(field_gradient, 4)) + ' T/m^2'
        self.Writer.DebugPrintout('\t' + debugstring)

    def Solenoid(self, linedict):
        soledata = linedict['data']
        length = soledata[0]  # First three non-blanks must be the entries in a specific order.
        field = soledata[1]   # Field in TRANSPORT units

        field_in_Gauss = field * self.Transport.scale[self.Transport.units['magnetic_fields'][0]]  # Convert to Gauss
        field_in_Tesla = field_in_Gauss * 1e-4  # Convert to Tesla

        lenInM = length * _General.ScaleToMeters(self.Transport, 'element_length')

        self.Transport.machineprops.solenoids += 1
        elementid = self._GetTransportElementName(linedict['name'])
        if not elementid:  # check on empty string
            elementid = 'SOLE'+_np.str(self.Transport.machineprops.solenoids)

        # pybdsim and pymadx are the same.
        self.Transport.machine.AddSolenoid(name=elementid, length=lenInM, ks=_np.round(field_in_Tesla, 4))

        self.Writer.DebugPrintout('\tConverted to:')
        debugstring = 'Solenoid ' + elementid + ', length ' + _np.str(lenInM) + \
                      ' m, ks ' + _np.str(_np.round(field_in_Tesla, 4)) + ' T'
        self.Writer.DebugPrintout('\t' + debugstring)

    def Printline(self, linedict):
        number = linedict['data'][0]
        self.Writer.DebugPrintout('\tTRANSPORT control line,')
        try:
            number = _np.float(number)
            if number == 48:
                self.Transport.machineprops.benddef = False
                self.Writer.DebugPrintout('\t48. Switched Dipoles to Angle definition.')
            elif number == 47:
                self.Transport.machineprops.benddef = True
                self.Writer.DebugPrintout('\t47. Switched Dipoles to field definition.')
            elif number == 19:
                if _General.CheckSingleLineOutputApplied(self.Transport.convprops.filename):
                    self.Transport.convprops.singleLineOptics = True
                self.Writer.DebugPrintout('\t19. Optics output switched to single line per element.')
            else:
                self.Writer.DebugPrintout('\tCode 13. ' + _np.str(number) + ' handling not implemented.')
        except ValueError:
            pass

    def Correction(self, linedict):
        if self.Transport.convprops.correctedbeamdef:
            self.Writer.DebugPrintout('\tNot Correction to original beam definition')
            return
        # Check if the previous line was the original beam definition and not an rms update
        if linedict['prevlinenum'] == 1.0 and not linedict['isAddition'] and self.Transport.convprops.beamdefined:
            self.Transport.convprops.correctedbeamdef = True

        correctiondata = linedict['data']
        if len(correctiondata) >= 15:  # 15 sigma elements
            sigma21 = correctiondata[0]
            sigma43 = correctiondata[5]
        else:
            self.Writer.DebugPrintout('\tLength of correction line is incorrect')
            return

        emittoverbeta = self.Transport.beamprops.SigmaXP**2 * (1 - sigma21**2)
        emittbeta = self.Transport.beamprops.SigmaX**2
        betx = _np.sqrt(emittbeta / emittoverbeta)
        emitx = emittbeta / betx
        slope = sigma21 * self.Transport.beamprops.SigmaXP / self.Transport.beamprops.SigmaX
        alfx = -1.0 * slope * betx

        self.Transport.beamprops.betx = betx
        self.Transport.beamprops.emitx = emitx / 1000.0
        self.Transport.beamprops.alfx = alfx

        emittoverbeta = self.Transport.beamprops.SigmaYP**2 * (1 - sigma43**2)
        emittbeta = self.Transport.beamprops.SigmaY**2
        bety = _np.sqrt(emittbeta / emittoverbeta)
        emity = emittbeta / bety
        slope = sigma43 * self.Transport.beamprops.SigmaYP / self.Transport.beamprops.SigmaY
        alfy = -1.0 * slope * bety

        self.Transport.beamprops.bety = bety
        self.Transport.beamprops.emity = emity / 1000.0
        self.Transport.beamprops.alfy = alfy

        self.Transport.beamprops.distrType = 'gausstwiss'

        self.Writer.DebugPrintout('\tConverted to:')
        self.Writer.DebugPrintout('\t Beam Correction. Sigma21 = ' + _np.str(sigma21) + ', Sigma43 = ' + _np.str(sigma43) + '.')
        self.Writer.DebugPrintout('\t Beam distribution type now switched to "gausstwiss":')
        self.Writer.DebugPrintout('\t Twiss Params:')
        self.Writer.DebugPrintout('\t BetaX = ' + _np.str(self.Transport.beamprops.betx) + ' ' + self.Transport.units['beta_func'])
        self.Writer.DebugPrintout('\t BetaY = ' + _np.str(self.Transport.beamprops.bety) + ' ' + self.Transport.units['beta_func'])
        self.Writer.DebugPrintout('\t AlphaX = ' + _np.str(self.Transport.beamprops.alfx))
        self.Writer.DebugPrintout('\t AlphaY = ' + _np.str(self.Transport.beamprops.alfy))
        self.Writer.DebugPrintout('\t Emittx = ' + _np.str(self.Transport.beamprops.emitx) + ' ' + self.Transport.units['emittance'])
        self.Writer.DebugPrintout('\t EmittY = ' + _np.str(self.Transport.beamprops.emity) + ' ' + self.Transport.units['emittance'])

    def SpecialInput(self, linedict):
        specialdata = linedict['data']
        self.Writer.DebugPrintout('\tSpecial Input line:')

        if specialdata[0] == 5.0:  # beampiperadius (technically only vertical, but will apply a circle for now)
            self.Writer.DebugPrintout('\tType 5: Vertical half aperture,')
            self.Transport.machineprops.dipoleVertAper = specialdata[1]
            self.Writer.DebugPrintout('\tHalf aperture set to ' + _np.str(specialdata[1]) + '.')
            if self.Transport.machineprops.fringeIntegral == 0:
                self.Transport.machineprops.fringeIntegral = 0.5  # default if a vertical aperture is specified.
                self.Writer.DebugPrintout('Fringe field integral not set, setting to default of 0.5.')
        elif specialdata[0] == 7.0:  # Fringe Field integral
            self.Transport.machineprops.fringeIntegral = specialdata[1]
            self.Writer.DebugPrintout('\tType 7: K1 Fringe field integral,')
            self.Writer.DebugPrintout('\tIntegral set to ' + _np.str(specialdata[1]) + '.')
        elif specialdata[0] == 8.0:  # Second Fringe Field integral
            self.Transport.machineprops.secondfringeInt = specialdata[1]
            self.Writer.DebugPrintout('\tType 7: K2 Fringe field integral,')
            self.Writer.DebugPrintout('\tIntegral set to ' + _np.str(specialdata[1]) + '.')
        elif specialdata[0] == 14.0:  # Definition of element type code 6.
            if self.Transport.convprops.typeCode6IsTransUpdate:
                self.Transport.convprops.typeCode6IsTransUpdate = False
                typeCode6def = 'Collimator'
            else:
                self.Transport.convprops.typeCode6IsTransUpdate = True
                typeCode6def = 'Transform Update'
            self.Writer.DebugPrintout('\tType 14: Type code 6 definition,')
            self.Writer.DebugPrintout('\tDefinition set to ' + typeCode6def + '.')
        elif specialdata[0] == 12.0:  # entrance poleface curvature
            self.Transport.machineprops.bendInCurvature = specialdata[1]
            self.Writer.DebugPrintout('\tType 12: Dipole entrance poleface curvature,')
            self.Writer.DebugPrintout('\tCurvature set to ' + _np.str(specialdata[1]) + '.')
        elif specialdata[0] == 13.0:  # exit poleface curvature
            self.Transport.machineprops.bendOutCurvature = specialdata[1]
            self.Writer.DebugPrintout('\tType 13: Dipole exit poleface curvature,')
            self.Writer.DebugPrintout('\tCurvature set to ' + _np.str(specialdata[1]) + '.')
        elif specialdata[0] == 16.0:  # X0 offset
            self.Transport.beamprops.X0 = specialdata[1]
            self.Writer.DebugPrintout('\tType 16: X0 beam offset,')
            self.Writer.DebugPrintout('\tOffset set to ' + _np.str(specialdata[1]) + '.')
        elif specialdata[0] == 17.0:  # Y0 offset
            self.Transport.beamprops.Y0 = specialdata[1]
            self.Writer.DebugPrintout('\tType 17: Y0 beam offset,')
            self.Writer.DebugPrintout('\tOffset set to ' + _np.str(specialdata[1]) + '.')
        elif specialdata[0] == 18.0:  # Z0 offset
            self.Transport.beamprops.Z0 = specialdata[1]
            self.Writer.DebugPrintout('\tType 18: Z0 beam offset,')
            self.Writer.DebugPrintout('\tOffset set to ' + _np.str(specialdata[1]) + '.')
        else:
            self.Writer.DebugPrintout('\tCode type not yet supported, or unknown code type.')

    def UnitChange(self, linedict):
        """
        Function to change the units (scaling) of various parameters.
        """
        label = linedict['label']
        number = linedict['number']
        if label == 'CM' or label == 'MM' or label == 'UM' or label == 'NM':
            label = label.lower()
        # Convert Energy Unit Cases:
        if label == 'EV':
            label = 'eV'
        if label == 'KEV':
            label = 'keV'
        if label == 'MEV':
            label = 'MeV'
        if label == 'GEV':
            label = 'GeV'
        if label == 'TEV':
            label = 'TeV'

        debugstring2 = '\tConverted to ' + label

        if _np.float(number) == 1:  # Horizontal and vertical beam size
            self.Transport.units['x'] = label
            self.Transport.units['y'] = label
            self.Transport.units['bend_vert_gap'] = label
            # self.units['pipe_rad'] = label
            debugstring1 = '\tType 1: Horizontal and vertical beam extents, and magnet apertures,'

        elif _np.float(number) == 2:  # Horizontal and vertical divergence
            self.Transport.units['xp'] = label
            self.Transport.units['yp'] = label
            debugstring1 = '\tType 2: Horizontal and vertical angles,'

        elif _np.float(number) == 3:  # Bending Magnet Gap
            self.Transport.units['y'] = label
            self.Transport.units['bend_vert_gap'] = label
            debugstring1 = '\tType 3: Vertical (only) beam extent and magnet aperture,'

        elif _np.float(number) == 4:  # Vertical Divergence ONLY
            self.Transport.units['yp'] = label
            debugstring1 = '\tType 4: Vertical (only) beam angle,'

        elif _np.float(number) == 5:  # Pulsed Beam Length
            self.Transport.units['bunch_length'] = label
            debugstring1 = '\tType 5: Bunch length,'

        elif _np.float(number) == 6:  # Momentum Spread
            self.Transport.units['momentum_spread'] = label  # Percent
            debugstring1 = '\tType 6: Momentum spread,'

        elif _np.float(number) == 7:  # Bend/pole face rotation
            debugstring1 = '\tType 7: Bend and poleface rotation angles,'
            debugstring2 = '\tCONVERTION NOT IMPLEMENTED YET.'
            pass

        elif _np.float(number) == 8:  # Element Length
            self.Transport.units['element_length'] = label
            debugstring1 = '\tType 8: Element length,'

        elif _np.float(number) == 9:  # Magnetic Field
            self.Transport.units['magnetic_fields'] = label
            debugstring1 = '\tType 9: Magnetic Fields,'

        elif _np.float(number) == 10:  # Mass
            debugstring1 = '\tType 10: Mass,'
            debugstring2 = '\tCONVERTION NOT IMPLEMENTED YET.'
            pass

        elif _np.float(number) == 11:  # Momentum / energy gain during acc.
            self.Transport.units['p_egain'] = label
            debugstring1 = '\tType 11: Momentum and accelerator energy gain,'
        else:
            # default output
            debugstring1 = '\tCode type not yet supported, or unknown code type.'
            debugstring2 = ''

        self.Writer.DebugPrintout('\tUnit change line:')
        self.Writer.DebugPrintout(debugstring1)
        self.Writer.DebugPrintout(debugstring2)

    def TransformUpdate(self, linedict):
        if linedict['elementnum'] == 6.0:
            errorline = '\tElement is either a transform update or a collimator. The type code 6 definition'
            errorline2 = '\thas not been switched to collimators, therefore nothing will be done for this element.'
            self.Writer.DebugPrintout(errorline)
            self.Writer.DebugPrintout(errorline2)

    def _GetTransportElementName(self, elementName):
        """
        Checks if name already used by an element in the machine.
        Appends a _N to the name where N is the lowest integer not already
        used in a elementName_N name in the machine.
        """
        if not self.Transport.convprops.keepName:
            return ''

        currElements = self.Transport.machine.elementsd.keys()
        updatedName = elementName
        if elementName in currElements:
            nameNumber = 1
            while True:
                updatedName = elementName + "_" + _np.str(nameNumber)
                if updatedName not in currElements:
                    break
                else:
                    nameNumber += 1

        return updatedName

    def _UpdateElementsFromFits(self):
        # Functions that update the elements in the element registry.
        # For debugging purposes, they return dictionaries of the element type,
        # length change details, and which parameters were updated and the values in a list which
        # follows the pattern of [parameter name (e.g. 'field'),oldvalue,newvalue]

        # Length update common to nearly all elements, seperate function to prevent duplication
        # fitIndex was used in the past. Pass in anyway in case of future need, just delete for now.
        def _updateLength(eleIndex, fitIndex, element):
            del fitIndex
            oldlength = self.Transport.ElementRegistry.elements[eleIndex]['length']
            lengthDiff = self.Transport.ElementRegistry.elements[eleIndex]['length'] - element['length']
            self.Transport.ElementRegistry.elements[eleIndex]['length'] = element['length']  # Update length
            self.Transport.ElementRegistry.length[eleIndex:] += lengthDiff  # Update running length of subsequent elements.
            self.Transport.ElementRegistry._totalLength += lengthDiff  # Update total length
            lendict = {'old': _np.round(oldlength, 5),
                       'new': _np.round(element['length'], 5)}
            return lendict

        def _updateDrift(eleIndex, fitIndex, element):
            eleDict = {'updated': False,
                       'element': 'Drift',
                       'params': []}

            # Only length can be varied
            if self.Transport.ElementRegistry.elements[eleIndex]['length'] != element['length']:
                lendict = _updateLength(eleIndex, fitIndex, element)
                eleDict['updated'] = True
                eleDict['length'] = lendict
            return eleDict

        def _updateQuad(eleIndex, fitIndex, element):
            eleDict = {'updated': False,
                       'element': 'Quadrupole',
                       'params': []}

            if self.Transport.ElementRegistry.elements[eleIndex]['data'][1] != element['data'][1]:  # Field
                oldvalue = self.Transport.ElementRegistry.elements[eleIndex]['data'][1]
                self.Transport.ElementRegistry.elements[eleIndex]['data'][1] = element['data'][1]
                eleDict['updated'] = True
                data = ['field', oldvalue, element['data'][1]]
                eleDict['params'].append(data)


            if self.Transport.ElementRegistry.elements[eleIndex]['length'] != element['length']:
                self.Transport.ElementRegistry.elements[eleIndex]['data'][0] = element['data'][0]  # Length in data
                lendict = _updateLength(eleIndex, fitIndex, element)
                eleDict['length'] = lendict
                eleDict['updated'] = True

            return eleDict

        def _updateDipole(eleIndex, fitIndex, element):
            eleDict = {'updated': False,
                       'element': 'Dipole',
                       'params': []}

            # TODO: Need code in here to handle variation in poleface rotation. Not urgent for now.
            if self.Transport.ElementRegistry.elements[eleIndex]['data'][1] != element['data'][1]:  # Field
                oldvalue = self.Transport.ElementRegistry.elements[eleIndex]['data'][1]
                if not self.Transport.machineprops.benddef:  # Transport can switch dipole input definition
                    par = 'field'
                    self.Transport.ElementRegistry.elements[eleIndex]['data'][1] = element['data'][1]
                    eleDict['updated'] = True
                else:
                    par = 'angle'
                    self.Transport.ElementRegistry.elements[eleIndex]['data'][1] = element['data'][3]
                    eleDict['updated'] = True
                data = [par, oldvalue, element['data'][3]]
                eleDict['params'].append(data)
            if self.Transport.ElementRegistry.elements[eleIndex]['length'] != element['length']:
                self.Transport.ElementRegistry.elements[eleIndex]['data'][0] = element['data'][0]  # Length in data
                lendict = _updateLength(eleIndex, fitIndex, element)
                eleDict['updated'] = True
                eleDict['length'] = lendict
            return eleDict

        self.Writer.DebugPrintout("")
        self.Writer.DebugPrintout("Updating elements with results from fitting")
        for index, name in enumerate(self.Transport.FitRegistry._uniquenames):
            fitindex = self.Transport.FitRegistry.GetElementIndex(name)
            eleindex = self.Transport.ElementRegistry.GetElementIndex(name)

            for fitnum, fit in enumerate(fitindex):
                for elenum, ele in enumerate(eleindex):
                    fitelement = self.Transport.FitRegistry.elements[fitindex[fitnum]]
                    if fitelement['elementnum'] == 3:
                        eledict = _updateDrift(eleindex[elenum], fitindex[fitnum], fitelement)
                    elif fitelement['elementnum'] == 4:
                        eledict = _updateDipole(eleindex[elenum], fitindex[fitnum], fitelement)
                    elif fitelement['elementnum'] == 5:
                        eledict = _updateQuad(eleindex[elenum], fitindex[fitnum], fitelement)

                    if eledict['updated']:
                        self.Writer.DebugPrintout(
                            "Element " + _np.str(eleindex[elenum]) + " was updated from fitting:")
                        self.Writer.DebugPrintout("\tOptics Output line:")
                        self.Writer.DebugPrintout(
                            "\t\t'" + self.Transport.FitRegistry.lines[fitindex[fitnum]] + "'")
                        if eledict.has_key('length'):
                            lenline = "\t" + eledict['element'] + " length updated to "
                            lenline += _np.str(eledict['length']['new'])
                            lenline += " (from " + _np.str(eledict['length']['old']) + ")."
                            self.Writer.DebugPrintout(lenline)
                        for param in eledict['params']:
                            parline = "\t" + eledict['element'] + " " + param[0]
                            parline += " updated to " + _np.str(param[2]) + " (from " + _np.str(param[1]) + ")."
                            self.Writer.DebugPrintout(parline)
                            # self.Writer.DebugPrintout("\n")

                        break



# pytransport.Reader - Transport input and output file readers.
# Version 1.0
# W. Shields and J. Snuverink
# william.shields.2010@live.rhul.ac.uk

"""
Reader

Readers for loading Transport input files and output files.
Can extract the individual lattice, optics, and fitting sections.

Classes:
Reader - a list of data read from Transport files.
ConversionData - a class for holding data during conversion.

"""

import numpy as _np
from Data import BDSData as _BDA


_allowedIndicatorLines = ['0  100', '0    0']

def GetOptics(inputFile, inputType=None):
    """
    Extract the optics from a Transport output file.
    """
    optics = _Optics()  # Instantiate empty data optics container

    if isinstance(inputType, _np.str):
        if inputType == 'beam':
            transdata = optics._getBeamOptics(inputFile)
            return transdata
        elif inputType == 'standard':
            transdata = optics._getStandardOptics(inputFile)
            return transdata

    f = open(inputFile)
    transdata = None
    for line in f:
        # remove any carriage returns (both Mac and Unix)
        line = line.rstrip('\r\n')
        splitline = _remove_blanks(line.split(' '))
        if splitline and splitline[0] == '*BEAM*':  # Is beam output
            # print "'*BEAM*' found in line " + _np.str(i+1)
            f.close()
            transdata = optics._getBeamOptics(inputFile)
            break
        elif line in _allowedIndicatorLines:
            # print "'0    0' found in line " + _np.str(i+1)
            f.close()
            transdata = optics._getStandardOptics(inputFile)
            break
        else:
            pass
    if transdata is None:
        errorstring = "Could not find an indicator in the file for either a beam output file\n"
        errorstring += "(indicator = '*BEAM*), or a standard output file (indicator = '0    0').\n"
        errorstring += "Please check the input file or specify the input type with the type argument \n"
        errorstring += "in the get_output function. Note that the only accepted values for type are \n"
        errorstring += "'standard' or 'beam'."
        raise IOError(errorstring)

    #transdata.MergeDuplicatesAtSameS()

    return transdata

def GetLattice(inputFile):
    """
    Function to extract the lattice from a standard output file.
    """
    flist = _LoadFile(inputFile)
    latticestart = 0
    latticeend = 0
    foundlatticestart = False
    foundlatticeend = False
    lattice = ['OUTPUT LATTICE']
    for linenum, line in enumerate(flist):
        if line in _allowedIndicatorLines:
            if not foundlatticestart:
                latticestart = linenum+1
                if flist[linenum+1] == '0INDICATOR VALUE WRONG OR MISSING - ZERO ASSUMED':
                    latticestart += 1
                foundlatticestart = True
        if line == '0SENTINEL':
            if not foundlatticeend:
                latticeend = linenum
                foundlatticeend = True
    if not foundlatticestart:
        if not foundlatticeend:
            raise IOError('No lattice found in ' + inputFile + '.')
        else:
            errorstring = 'The end of a lattice (line = "0SENTINEL") was found at line ' + _np.str(latticeend + 1) + ',\n'
            errorstring += 'but the start of a lattice (line = "0    0") was not found. Please check the input file.'
            raise IOError(errorstring)
    elif not foundlatticeend:
            errorstring = 'The start of a lattice (line = "0    0") was found at line ' + _np.str(latticestart - 1) + ',\n'
            errorstring += 'but the end of a lattice (line = "0SENTINEL") was not found. Please check the input file.'
            raise IOError(errorstring)
    else:
        lattice.extend(flist[latticestart:latticeend])
    return lattice

def GetFitsSection(inputFile):
    """
    Function to get the fit routine data from the standard transport output.
    Returns two lists, the first with the direct output from the fitting data,
    the second with the first line of each element in the output data, which contains the
    element parameters with their fitted values.
    """

    flist = _LoadFile(inputFile)
    fitstart = 0
    fitend = 0
    foundfitstart = False
    foundfitend = False
    fits = []
    for linenum, line in enumerate(flist):
        if not foundfitstart:
            if line == '0SENTINEL':
                fitstart = linenum
                foundfitstart = True
        if not foundfitend:
            splitline = _remove_blanks(line.split(' '))
            if splitline and splitline[0] == '*BEAM*' and not foundfitend:
                fitend = linenum
                foundfitend = True
    if not foundfitstart:
        print('No fitting output found.')
        return None
    elif foundfitstart and not foundfitend:
            errorstring = 'The start of the fitting output (first line containing "0SENTINEL") was found at line ' + _np.str(fitstart-1) + ',\n'
            errorstring += 'but the end of the fitting output (first line containing "*BEAM*") was not found. Please check the input file.'
            raise IOError(errorstring)
    fits.extend(flist[fitstart:fitend])
    return fits

def GetResultsFromFitting(inputFile):
    optics = _Optics() # Instantiate empty data optics container
    output = optics._getOptics(inputFile)
    fitres = [element[0] for element in output]
    return fitres


class _Optics:
    """
    Class for reading optics from Transport output files.
    The optics can be from standard output or Beam output.
    """
    def __init__(self):
        self.transdata = {
            'Sigma_x'   : [],
            'Sigma_xp'  : [],
            'Sigma_y'   : [],
            'Sigma_yp'  : [],
            'S'         : [],
            'Alpha_x'   : [],
            'Alpha_y'   : [],
            'Beta_x'    : [],
            'Beta_y'    : [],
            'Emitt_x'   : [],
            'Emitt_y'   : [],
            'Disp_x'    : [],
            'Disp_y'    : [],
            'Disp_xp'   : [],
            'Disp_yp'   : [],
            'Sigma_p'   : [],
            'Momentum'  : [],
            'E'         : [],  # kinetic energy
            'Name'      : [],
            'Type'      : []
            }
        # TODO: some unit for now, needs to be extracted from output / convert.py !
        self.transunits = {
            'Sigma_x'   : 'mm',
            'Sigma_xp'  : 'mrad',
            'Sigma_y'   : 'mm',
            'Sigma_yp'  : 'mrad',
            'S'         : 'm',
            'Alpha_x'   : '',
            'Alpha_y'   : '',
            'Beta_x'    : 'mm / mrad',
            'Beta_y'    : 'mm / mrad',
            'Emitt_x'   : 'mm mrad',
            'Emitt_y'   : 'mm mrad',
            'Disp_x'    : '',
            'Disp_y'    : '',
            'Disp_xp'   : '',
            'Disp_yp'   : '',
            'Sigma_p'   : 'MeV/c',
            'Momentum'  : 'MeV/c',
            'E'         : 'MeV',  # kinetic energy
            'Name'      : '',
            'Type'      : ''
            }

    def CheckSingleLineOutputApplied(self, inputfile):
        """
        Function to check if the control element that print element output in
        a single line was successfully applied. Check needed as not all versions
        of TRANSPORT can run this type code.
        """
        optics = self._getOptics(inputfile)
        for element in optics:
            if element == 'IO: UNDEFINED TYPE CODE 13. 19. ;':
                return True
        return False

    def _processBeamOptics(self, flist):
        """
        Process the optics from a Beam output file.
        """
        transdata = {
            'Sigma_x'   : [],
            'Sigma_xp'  : [],
            'Sigma_y'   : [],
            'Sigma_yp'  : [],
            'S'         : [],
            'Alpha_x'   : [],
            'Alpha_y'   : [],
            'Beta_x'    : [],
            'Beta_y'    : [],
            'Emitt_x'   : [],
            'Emitt_y'   : [],
            'Disp_x'    : [],
            'Disp_y'    : [],
            'Disp_xp'   : [],
            'Disp_yp'   : [],
            'Sigma_p'   : [],
            'Name'      : [],
            }
        num_elements = 0
        for elenum, element in enumerate(flist):
            if element == '':  # The first line of the section should be a blank line.
                try:
                    section = flist[elenum+1:elenum+13]

                    # Get the element name, type and start position (in s)
                    line0 = section[0]
                    line0split = line0.split('*')
                    # typestr = line0split[1]
                    reststr = line0split[2]
                    restsplit = reststr.split(' ')
                    restsplit = _remove_blanks(restsplit)
                    transdata['S'].append(_np.float(restsplit[2]))
                    try :
                        namestr = restsplit[4]
                    except IndexError:
                        namestr = ''
                    transdata['Name'].append(namestr)
                    # transdata['type'].append(typestr)

                    # Get sigma_x and sigma_xp
                    line3 = section[3].split(' ')
                    line3 = _remove_blanks(line3)
                    transdata['Sigma_x'].append(_np.float(line3[3])/1000)
                    transdata['Sigma_xp'].append(_np.float(line3[5])/1000)

                    # Get sigma_y and sigma_yp
                    line4 = section[4].split(' ')
                    line4 = _remove_blanks(line4)
                    transdata['Sigma_y'].append(_np.float(line4[3])/1000)
                    transdata['Sigma_yp'].append(_np.float(line4[5])/1000)
                    
                    # Get momentum spread
                    line5 = section[5].split(' ')
                    transdata['Sigma_p'].append(_np.float(line5[5])/100)
                    
                    # Get alfa and beta twiss transdata for x and y.
                    line7 = section[7].split(' ')
                    line7 = _remove_blanks(line7)
                    transdata['Alpha_x'].append(_np.float(line7[0]))
                    transdata['Beta_x'].append(_np.float(line7[1]))
                    transdata['Alpha_y'].append(_np.float(line7[3]))
                    transdata['Beta_y'].append(_np.float(line7[4]))

                    # Get horizontal and vertical dispersion
                    line10 = section[10].split(' ')
                    line10 = _remove_blanks(line10)
                    transdata['Disp_x'].append(_np.float(line10[2])/10)
                    transdata['Disp_y'].append(_np.float(line10[5])/10)
                    
                    # Terms for calculating the emittance.
                    term1x = _np.float(line3[3])**2
                    term2x = (_np.float(line10[2])*(_np.float(line5[5])/100))**2
                    term1y = _np.float(line4[3])**2
                    term2y = (_np.float(line10[5])*(_np.float(line5[5])/100))**2

                    # Get horizontal and vertical emittance
                    emittx = (term1x - term2x) / _np.float(line7[1])
                    emitty = (term1y - term2y) / _np.float(line7[4])
                    transdata['Emitt_x'].append(emittx)
                    transdata['Emitt_y'].append(emitty)
                
                    num_elements += 1
                except ValueError:
                    errstr = "Could not process section beginning at line " + _np.str(elenum) + " : "
                    print(errstr)
                    print " "
                    print(element)
            elif element == "EOF -- rewind file":
                break

        data = _BDA()  # Now convert the dict into BDSData instance for final output.
        for keyName in transdata.keys():
            data._AddProperty(keyName)
        for i in range(num_elements):
            data.append(_GetElementData(i, transdata))

        return data

    def _getStandardOptics(self, inputFile):
        """
        Get the optics from a standard output file. Returns a pytransport.Data.BDSData object.
        """
        if self.CheckSingleLineOutputApplied(inputFile):
            optics = self._processStandardOpticsSingleLine(inputFile)
        else:
            optics = self._processStandardOpticsMultiLines(inputFile)
        return optics

    def _processStandardOpticsMultiLines(self, inputFile):
        """
        Process the optics from a standard output file when written to multiple lines.
        """
        elementlist = self._getOptics(inputFile)
        notokElements = ['AXIS SHIFT', 'FIT']
        
        num_elements = 0
        # initialise momentum/energy since not given for every element
        momentum = 0.0
        energy = 0.0
        proton_mass = 938.272

        # cumulative machine length for s position. S position in optics output is rounded, too inaccurate.
        length = 0

        for element in elementlist:
            if (not isinstance(element, _np.str)) and (len(element) > 1):  # I.e not a fit or matrix-modifying element
                # type is in between * can have a space (for space charge *SP CH*)
                elementType = element[0].split('*')[1]
                elementProperties = _remove_blanks(element[0].split('*')[2].split(' '))
                elementSigmaMatrix = element[1:7]
                elementTransMatrix = element[8:]

                if elementType not in notokElements:
                    elename = elementProperties[1].strip('"')
                    if elementType == "BEAM" or elementType == "ACC":
                        momentum = _np.float(elementProperties[-2])
                        energy = _np.sqrt(proton_mass*proton_mass + momentum*momentum) - proton_mass

                    # sigmax line may or may not have element coordinates too, so cannot assume index counting from 0
                    firstLine = _remove_blanks(elementSigmaMatrix[0].split(' '))
                    try:
                        sigx = _np.float(firstLine[-2])
                    except ValueError:
                        sigx = _np.float(_remove_blanks(elementSigmaMatrix[0].split(' '))[3])

                    if '*COORDINATES*' in firstLine:
                        x = _np.float(firstLine[3])
                        y = _np.float(firstLine[4])
                        z = _np.float(firstLine[5])

                    sigxp   = _np.float(_remove_blanks(elementSigmaMatrix[1].split(' '))[1])
                    sigy    = _np.float(_remove_blanks(elementSigmaMatrix[2].split(' '))[1])
                    sigyp   = _np.float(_remove_blanks(elementSigmaMatrix[3].split(' '))[1])
                    sigt    = _np.float(_remove_blanks(elementSigmaMatrix[4].split(' '))[1])
                    sigp    = _np.float(_remove_blanks(elementSigmaMatrix[5].split(' '))[1])
                    r21     = _np.float(_remove_blanks(elementSigmaMatrix[1].split(' '))[3])
                    r43     = _np.float(_remove_blanks(elementSigmaMatrix[3].split(' '))[5])

                    dx      = _np.float(_remove_blanks(elementSigmaMatrix[5].split(' '))[3])
                    dxp     = _np.float(_remove_blanks(elementSigmaMatrix[5].split(' '))[4])
                    dy      = _np.float(_remove_blanks(elementSigmaMatrix[5].split(' '))[5])
                    dyp     = _np.float(_remove_blanks(elementSigmaMatrix[5].split(' '))[6])

                    #dx = _GetTransformLineElements(element[8])[5]
                    #dxp = _GetTransformLineElements(element[9])[5]
                    #dy = _GetTransformLineElements(element[10])[5]
                    #dyp = _GetTransformLineElements(element[11])[5]

                    # get s position after updating machine length with element length
                    # transport output is column aligned, check the parameter has unit of M.
                    # has to be "M  ", could be "MEV" in beam element.
                    lengthStr = element[0][42:55]
                    if lengthStr[-3:] == "M  ":
                        length += _np.float(lengthStr[:-3])
                    s = length

                    self._SetTransportData(sigx, sigxp, sigy, sigyp, s, dx, dy, dxp, dyp, sigp, momentum, energy,
                                           elename, elementType, r21, r43)
                    num_elements += 1

        data = _BDA()      # Now convert the dict into BDSData instance for final output.
        for keyName, unit in self.transunits.iteritems():
            data._AddProperty(keyName, unit)
        for i in range(num_elements):
            data.append(_GetElementData(i, self.transdata))

        return data

    def _SetTransportData(self, sigx, sigxp, sigy, sigyp, s, dx, dy, dxp, dyp, sigp, momentum, energy, elename, elementType,
                          r21, r43):
        """
        Set the beam data.
        """
        # Add/Subtract small amount if sin of phase space ellipse rotation is +/-one.
        # This comes from the output annoyingly rounding the code to one ,
        # which produces a div by zero later in the beta and gamma calculations.
        if r21 == 1.0:
            r21 -= 1e-4
        if r21 == -1.0:
            r21 += 1e-4

        if r43 == 1.0:
            r43 -= 1e-4
        if r43 == -1.0:
            r43 += 1e-4

        # Calculate twiss parameters
        xpint = _np.sqrt(sigxp ** 2 * (1 - r21 ** 2))
        ypint = _np.sqrt(sigyp ** 2 * (1 - r43 ** 2))

        ex = sigx * xpint
        ey = sigy * ypint

        if ex is 0:
            betx = 0
        else:
            betx = (sigx ** 2.0) / ex

        if ey is 0:
            bety = 0
        else:
            bety = (sigy ** 2.0) / ey

        if _np.isnan(betx):
            betx = 0
        if _np.isnan(bety):
            bety = 0

        alfx = -betx * r21 * sigxp/sigx
        alfy = -bety * r43 * sigyp/sigy

        self.transdata['Sigma_x'].append(sigx / 1000)  # convert to m
        self.transdata['Sigma_xp'].append(sigxp / 1000)  # convert to rad
        self.transdata['Sigma_y'].append(sigy / 1000)  # convert to m
        self.transdata['Sigma_yp'].append(sigyp / 1000)  # convert to rad
        self.transdata['S'].append(s)
        self.transdata['Alpha_x'].append(alfx)
        self.transdata['Alpha_y'].append(alfy)
        self.transdata['Beta_x'].append(betx)
        self.transdata['Beta_y'].append(bety)
        self.transdata['Emitt_x'].append(ex)
        self.transdata['Emitt_y'].append(ey)
        self.transdata['Disp_x'].append(dx)
        self.transdata['Disp_y'].append(dy)
        self.transdata['Disp_xp'].append(dxp)
        self.transdata['Disp_yp'].append(dyp)
        self.transdata['Sigma_p'].append(sigp)
        self.transdata['Momentum'].append(momentum)
        self.transdata['E'].append(energy)
        self.transdata['Name'].append(elename)
        self.transdata['Type'].append(elementType)

    def _processStandardOpticsSingleLine(self, inputFile):
        """
        Process the optics from a standard output file when written to single lines as specified
        by a 13. 19. element in Transport.
        """
        elementlist = self._getOptics(inputFile)
        # seperate R matrix table from sigma matrix elements
        rMatrixElements = elementlist[-1]
        rMatrix = []

        # Second to last is column headers for R matrix table
        sMatrix = elementlist[:-2]

        for element in rMatrixElements[1:]:
            rMatrix.append(_remove_blanks(element.split(' ')))

        num_elements = 0
        momentum = 0.0
        energy = 0.0
        proton_mass = 938.272
        notokElements = ['AXIS SHIFT']
        okRElements = [3, 4, 5]  # ok element types for R matrix matching

        for element in sMatrix:
            if len(element) > 1:  # I.e not a fit or matrix-modifying element
                elementLine = _remove_blanks(element[0].split(' '))
                elementLine = _updateElementLine(elementLine)
                elementType = elementLine[0].strip('*')  # element type
                # typenum = _np.float(elementLine[1])

                # Get line with sigma data.
                # Element can have an additional line with fitting vary code data
                if len(element) == 3:
                    sigmaLine = element[2]
                else:
                    sigmaLine = element[1]

                if elementType not in notokElements:
                    elename = _removeIllegals(elementLine[2])  # remove illegal characters

                    if elementType == "BEAM" or elementType == "ACC":
                        momentum = _np.float(elementLine[-2])
                        energy = _np.sqrt(proton_mass*proton_mass + momentum*momentum) - proton_mass

                    if len(element) > 6:  # In case beam is defined before output format change.
                        s     = _np.float(_remove_blanks(element[1].split(' '))[0])
                        sigx  = _np.float(_remove_blanks(element[1].split(' '))[3])
                        sigxp = _np.float(_remove_blanks(element[2].split(' '))[1])
                        sigy  = _np.float(_remove_blanks(element[3].split(' '))[1])
                        sigyp = _np.float(_remove_blanks(element[4].split(' '))[1])
                        sigt  = _np.float(_remove_blanks(element[5].split(' '))[1])
                        sigp  = _np.float(_remove_blanks(element[6].split(' '))[1])

                        try:
                            r21 = _np.float(_remove_blanks(element[2].split(' '))[3])
                        except IndexError:
                            r21 = 0
                        try:
                            r43 = _np.float(_remove_blanks(element[4].split(' '))[5])
                        except IndexError:
                            r43 = 0
                    else:
                        s     = _np.float(_remove_blanks(sigmaLine.split(' '))[0])
                        sigx  = _np.float(_remove_blanks(sigmaLine.split(' '))[2])
                        sigxp = _np.float(_remove_blanks(sigmaLine.split(' '))[4])
                        sigy  = _np.float(_remove_blanks(sigmaLine.split(' '))[6])
                        sigyp = _np.float(_remove_blanks(sigmaLine.split(' '))[8])
                        sigt  = _np.float(_remove_blanks(sigmaLine.split(' '))[10])
                        sigp  = _np.float(_remove_blanks(sigmaLine.split(' '))[12])
                        r21   = _np.float(_remove_blanks(sigmaLine.split(' '))[14])
                        r43   = _np.float(_remove_blanks(sigmaLine.split(' '))[15])

                    dx = 0
                    dy = 0
                    dxp = 0
                    dyp = 0
                    # TODO: Look up where dxp and dyp are in the output. Leave as zero for now.
                    # Find matching R matrix element and get dispersion
                    for rElement in rMatrix:
                        if _np.float(rElement[1]) in okRElements and (_np.float(rElement[0]) == s) \
                                and (rElement[2] == elename):
                            # Dispersion position dependent on existence of field strength in output
                            # Field strength written before first *, which should be the 4th element
                            if rElement.index('*') == 4:
                                dx = _np.float(rElement[15])
                                dy = _np.float(rElement[17])
                            else:
                                dx = _np.float(rElement[14])
                                dy = _np.float(rElement[16])

                    self._SetTransportData(sigx, sigxp, sigy, sigyp, s, dx, dy, dxp, dyp, sigp, momentum, energy,
                                           elename, elementType, r21, r43)
                    num_elements += 1

        data = _BDA()      # Now convert the dict into BDSData instance for final output.
        for keyName, unit in self.transunits.iteritems():
            data._AddProperty(keyName, unit)
        for i in range(num_elements):
            data.append(_GetElementData(i, self.transdata))
        
        return data

    def _getOptics(self, filename):
        """
        Function to extract the output from a standard output file. The output will be an list of the lines
        for each element which contains the beam data. Each element should contain the R and TRANSPORT matrices
        which are necessary so the beam info can be calculated.
        """
        flist = _LoadFile(filename)
        foundOpticsStart = False
        foundOpticsEnd = False
        foundRMatrixElementStart = False

        for linenum, line in enumerate(flist):
            splitline = _remove_blanks(line.split(' '))
            if splitline and splitline[0] == '*BEAM*' and not foundOpticsStart:
                opticsStart = linenum
                foundOpticsStart = True
            if splitline and splitline[0] == '0*LENGTH*':
                opticsEnd = linenum
                foundOpticsEnd = True
            if splitline and splitline[0] == '0POSITION':
                rMatrixElementStart = linenum
                foundRMatrixElementStart = True
    
        if not foundOpticsStart:
            if not foundOpticsEnd:
                raise IOError('No output found in ' + filename + '.')
            else:
                errorstring = 'The end of a lattice (line containing "0*LENGTH*") was found at ' \
                              'line ' + _np.str(opticsStart + 1)+',\n'
                errorstring += 'but the start of a lattice (first line containing "*BEAM*") was not found. ' \
                               'Please check the input file.'
                raise IOError(errorstring)
        elif not foundOpticsEnd:
                errorstring = 'The start of a lattice (first line containing "*BEAM*") was found at ' \
                              'line ' + _np.str(opticsStart - 1)+',\n'
                errorstring += 'but the end of a lattice (line containing "0*LENGTH*") was not found. ' \
                               'Please check the input file.'
                raise IOError(errorstring)
        else:
            output = flist[opticsStart:opticsEnd]

        # Append rest of the file which should only contain a table of R Matrix elements.
        if foundRMatrixElementStart:
            output.extend(flist[rMatrixElementStart:])

        # Split the list of all element data into their individual elements.
        elementlist = []
        elementstart = False
        elementend = False
        finalelement = False
        for linenum, line in enumerate(output):
            if linenum == (len(output)-1):
                finalelement = True
            try:
                if (line[1] == '*') or (line[:9] == '0POSITION'):
                    if line[2:11] == 'TRANSFORM':  # Is midway through element output
                        pass
                    elif elementstart is False:  # Current line must be start of the element
                        elementstart = True
                        elementstartline = linenum
                    elif elementstart is True:
                        elementend = True       # Otherwise the line must be the start of the next element
                        elementendline = linenum
                if elementstart and finalelement:
                    element = output[elementstartline:]
                    if element[-1][:2] == 'IO':
                        elementlist.append(element[:-1])
                        elementlist.append(element[-1])
                    else:
                        elementlist.append(element)
                if elementstart and elementend:  # If the start and end of the element are found, append and reset
                    element = output[elementstartline:elementendline]
                    if element[-1][:2] == 'IO':
                        elementlist.append(element[:-1])
                        elementlist.append(element[-1])
                    else:
                        elementlist.append(element)
                    elementend = False
                    elementstart = False
                if elementstart is False and elementend is False:  # Though if it's been reset, it must be because the
                    elementstart = True                            # current line is the start of next element, so set
                    elementstartline = linenum                     # the start line for the next element
            except IndexError:
                pass
        return elementlist

    def _getBeamOptics(self, inputFile):
        """
        Returns a BDSData instance of parameters from the input file.
        The input file is assumed to contain the beam data as output
        manually by TRANSPORT. As such, the structure of said output
        is assumed to remain unchanged from lattice to lattice. This code
        is designed to work with that output only. An example of that
        output would be:

        *DRIFT*      z = 34.372 m
        *SIGMA*
         Center:         0.000 mm    0.000 mrad    0.000 mm    0.000 mrad
         horz. Par. :   18.361 mm   23.134 mrad    0.997
         vert. Par. :    8.447 mm   10.242 mrad   -0.995
        *TWISS PARAMETERS* (for dp/p = 1.000 % )
           alfax:     betax:         alfay:     betay:
         -13.51800   10.75807 m      9.55361    7.92270 m
        *TRANSFORM 1*
            horz:                              vert:
          -14.22000   -0.64752   -9.32539      0.13646    0.97529    0.00000
          -18.40864   -0.90858  -10.33185     -1.18959   -1.17399    0.00000

        Anything other than this format will be read incorrectly or will
        not be read. There has to be a blank line between every element's
        output in the file.

        Note:
            It is assumed that the element (3,2) in the displayed *TRANSFORM 1*
            matrices corresponds to the dispersion (-9.32539 and 0.0000 in the above
            example). Some output however appears to have the correct magnitude, but
            incorrect sign. This doesn't affect the resulting beam size, but beware
            that a direct dispersion comparison to another lattice may appear incorrect.
        """
        flist = _LoadFile(inputFile)
        transdata = self._processBeamOptics(flist)
        return transdata


def _remove_blanks(line):
    """
    Removes any blanks from a string (blanks being '' and not white spaces).
    """
    newline = ''
    for element in line:
        if element != '':
            newline += element
            newline += ' '
    stripline = newline.split(' ')
    # remove last element as it will be blank (due to added space)
    return stripline[:-1]


def _removeIllegals(line):
    """
    Function to remove '' and stray characters from lines.
    """
    illegal = ['"', '', '(', ')']

    modLine = ''
    for element in line:
        if element not in illegal:
            modLine += element
    return modLine


def _split_negatives(line):
    newline = []
    for element in line:
        negpos = element.find('-')
        if negpos > 1:
            parts = element.split('-')
            newline.append(parts[0])
            newline.append('-' + parts[1])
        else:
            newline.append(element)
    return newline


def _LoadFile(inputfile):
    """
    Converts the input file into a list. The data has to be in a format other than
    handling line by line (using next()).

    The function for processing the file reads section by section rather than line
    by line. This may be an inefficient method but the input file should not be very
    large so it should not require a large amount of memory.
    """
    if inputfile == '':
        raise IOError('No file name supplied.')
    flist = []
    infile = open(inputfile)
    # Loop over lines and remove any carriage returns (both Mac and Unix)
    for line in infile:
        cleanline = line.rstrip('\r\n')
        flist.append(cleanline)
    infile.close()
    return flist


def _updateElementLine(line):
    if (line[0] == '*Z') and (line[1] == 'ROT*'):
        newline = []
        elementType = '*Z ROT*'
        newline.append(elementType)
        for element in line[2:]:
            newline.append(element)
    else:
        newline = line

    return newline


def _GetElementData(index, dataDict):
    # Function to get the data for each element, rather than each key.
    elementlist = [dataDict[keyName][index] for keyName in dataDict.keys()]
    return elementlist

def _GetTransformLineElements(line):
    elements = []
    for element in range(6):
        start = ((element + 1) * 10 + 1)
        end = start + 10
        eleVal = _remove_blanks(line[start:end].split(' '))[0]
        try:
            elements.append(_np.float(eleVal))
        except ValueError:
            elements.append(0)
    return elements


